"""Process the raw CMU-MOSEI dataset.

This assumes the file structure from the original compressed file:
/.../
    CMU_MOSEI_Labels.csd
    *.csd
    Raw/
        Audio/
            Full/
                WAV_16000/
                    *.wav
                COVAREP/
                    *.mat
        Videos/
            Full/
                Combined/
                    *.mp4
            Segmented/
                Combined/
                    *.mp4
    ...
"""

import json
from collections import defaultdict
from pathlib import Path
from typing import Dict

import click
import h5py
import numpy as np
import pandas as pd
import soundfile
from joblib import delayed
from splits import standard_test_fold, standard_train_fold, standard_valid_fold
from tqdm import tqdm

from ertk.dataset import write_annotations, write_filelist
from ertk.utils import TqdmParallel

emotion_map = {
    "happy": "happiness",
    "sad": "sadness",
}


def get_vote(vals):
    idx = vals.argmax()
    max = vals[idx]
    if max > vals.sum() / 2:
        # Majority
        return idx, "M"
    if np.count_nonzero(max == vals) == 1:
        # Plurality
        return idx, "P"
    return -1, "X"


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the CMU-MOSEI dataset at location INPUT_DIR and resample
    audio to 16 kHz 16-bit FLAC audio.
    """

    audio_dir = input_dir / "Raw" / "Audio" / "Full" / "WAV_16000"
    all_videos = sorted(set(x.stem for x in audio_dir.glob("*.wav")))

    transcripts_dir = input_dir / "Raw" / "Transcript" / "Segmented" / "Combined"
    transcripts = []
    for name in all_videos:
        trn_file = transcripts_dir / f"{name}.txt"
        with open(trn_file) as fid:
            for row in fid:
                # Some transcripts contain '_____' or '______'
                vals = row.strip().split("___", maxsplit=5)
                dtypes = [str, str, float, float, str]
                transcripts.append([f(x) for f, x in zip(dtypes, vals)])
    trn_df = pd.DataFrame(transcripts, columns=["video", "clip", "start", "end", "trn"])
    neg_starts = trn_df[trn_df["start"] < 0]
    if len(neg_starts) > 0:
        print(f"WARNING: {len(neg_starts)} clips with negative start value")
    trn_df.index = trn_df["video"] + "_" + trn_df["clip"]
    write_annotations(trn_df["trn"].to_dict(), "transcript")
    write_annotations(trn_df["video"].to_dict(), "video")

    dfs = []
    for csvfile in (input_dir / "Raw" / "Labels").glob("*.csv"):
        dfs.append(pd.read_csv(csvfile, dtype={15: str, 27: str, 28: str}))
    df = pd.concat(dfs)
    df["name"] = df["Input.VIDEO_ID"].str.cat(df["Input.CLIP"], sep="_")
    df["name"] = df["name"].map(lambda x: x.split("/")[1] if "/" in x else x)
    df["rater"] = df["WorkerId"]
    df.set_index(["name", "rater"], inplace=True)
    df.rename(
        columns={x: x[7:] for x in df.columns if x.startswith("Answer.")}, inplace=True
    )
    df = df[
        ["sentiment", "anger", "disgust", "fear", "happiness", "sadness", "surprise"]
    ]
    df.sort_index().to_csv("ratings.csv")

    def process(seg):
        start = max(0, int(16000 * trn_df.loc[seg, "start"]))
        end = max(0, int(16000 * trn_df.loc[seg, "end"]))
        if end == start:
            return
        base_clip = audio_dir / f"{trn_df.loc[seg, 'video']}.wav"
        audio, _ = soundfile.read(base_clip, start=start, stop=end)
        soundfile.write(
            resample_dir / f"{seg}.flac", audio, samplerate=16000, subtype="PCM_16"
        )

    resample_dir = Path("resampled")
    resample_dir.mkdir(exist_ok=True)
    if resample:
        TqdmParallel(len(trn_df.index), "Splitting audio", prefer="threads", n_jobs=-1)(
            delayed(process)(name) for name in trn_df.index
        )
    all_paths = list(resample_dir.glob("*.flac"))
    write_filelist(all_paths, "files_all")
    write_annotations({x.stem: "en" for x in all_paths}, "language")

    dataset = h5py.File(input_dir / "CMU_MOSEI_Labels.csd", "r")
    dim_names = json.loads(dataset["All Labels/metadata/dimension names"][0])
    labelled_videos = set(dataset["All Labels/data"].keys())

    labels_maj = {}
    labels_plu = {}
    dim_vals: Dict[str, Dict[str, float]] = defaultdict(dict)
    for name in tqdm(trn_df.index, "Processing labels"):
        video = trn_df.loc[name, "video"]
        if video in labelled_videos:
            features = dataset[f"All Labels/data/{video}/features"]
            intervals = dataset[f"All Labels/data/{video}/intervals"]
            # We need to match the clip IDs to the intervals
            for i, (start, end) in enumerate(intervals):
                if (
                    start == trn_df.loc[name, "start"]
                    and end == trn_df.loc[name, "end"]
                ):
                    max_emo_idx, tp = get_vote(features[i, 1:])
                    labels_maj[name] = (
                        dim_names[1:][max_emo_idx] if tp == "M" else "unknown"
                    )
                    labels_plu[name] = (
                        dim_names[1:][max_emo_idx] if tp in ["M", "P"] else "unknown"
                    )
                    for d_idx, dim in enumerate(dim_names):
                        dim_vals[dim][name] = features[i, d_idx]
                    break
    dataset.close()

    labels_maj = {k: emotion_map.get(v, v) for k, v in labels_maj.items()}
    labels_maj.update({k: "" for k in trn_df.index if k not in labels_maj})
    write_annotations(labels_maj, "label_maj")
    labels_plu = {k: emotion_map.get(v, v) for k, v in labels_plu.items()}
    labels_plu.update({k: "" for k in trn_df.index if k not in labels_plu})
    write_annotations(labels_plu, "label_plu")
    for dim in dim_names:
        dim_vals[dim].update({k: "" for k in trn_df.index if k not in dim_vals[dim]})
        write_annotations(dim_vals[dim], dim)
    labelled_paths = {resample_dir / f"{k}.flac" for k, v in labels_maj.items() if v}
    write_filelist(labelled_paths, "files_labels")

    fold_paths = defaultdict(list)
    vid_to_fold = {x: "train" for x in standard_train_fold}
    vid_to_fold.update({x: "valid" for x in standard_valid_fold})
    vid_to_fold.update({x: "test" for x in standard_test_fold})
    splits = {}
    for path in all_paths:
        video = path.stem.rsplit("_", maxsplit=1)[0]
        fold = vid_to_fold.get(video)
        splits[path.stem] = fold
        if fold is None:
            continue
        fold_paths[fold].append(path)
        if path in labelled_paths:
            fold_paths[f"{fold}_labels"].append(path)
    write_annotations(splits, "split")
    for key, pathlist in fold_paths.items():
        write_filelist(pathlist, f"files_{key}")


if __name__ == "__main__":
    main()



"""Audioset feature extractors."""

import logging
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, List, Union

import numpy as np
from omegaconf import MISSING

from ertk.config import ERTKConfig
from ertk.preprocessing.spectrogram import spectrogram
from ertk.utils import frame_array, is_mono_audio

from .._base import AudioClipProcessor, FeatureExtractor

__all__ = [
    "YamnetExtractor",
    "YamnetExtractorConfig",
    "VGGishExtractor",
    "VGGishExtractorConfig",
]


def _spectrogram(x: np.ndarray, sr: float) -> np.ndarray:
    return spectrogram(
        x,
        sr,
        kind="mel",
        pre_emphasis=0,
        window_size=0.025,
        window_shift=0.01,
        n_mels=64,
        fmin=125,
        fmax=7500,
        to_log="log",
        power=1,
    )


@dataclass
class YamnetExtractorConfig(ERTKConfig):
    """Configuration for a Yamnet extractor."""

    model_dir: str = MISSING
    """The directory containing the YAMNet model files."""


class YamnetExtractor(
    FeatureExtractor, AudioClipProcessor, fname="yamnet", config=YamnetExtractorConfig
):
    """A YAMNet extractor."""

    config: YamnetExtractorConfig

    def __init__(self, config: YamnetExtractorConfig) -> None:
        super().__init__(config)

        import tensorflow as tf

        from ertk.tensorflow.utils import init_gpu_memory_growth

        self._old_env_log_level = os.environ.get("TF_CPP_MIN_LOG_LEVEL")
        os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"  # Ignore INFO messages
        self._old_log_level = tf.get_logger().level
        tf.get_logger().setLevel(logging.WARNING)
        init_gpu_memory_growth()

        from ._yamnet.params import Params
        from ._yamnet.yamnet import yamnet

        input_layer = tf.keras.Input((96, 64))
        predictions, embeddings = yamnet(input_layer, Params())
        model = tf.keras.Model(inputs=input_layer, outputs=[predictions, embeddings])
        model.load_weights(str(Path(config.model_dir, "yamnet.h5")))
        model(tf.zeros((1, 96, 64)))
        self.model = model

    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        if is_mono_audio(x):
            x = _spectrogram(x, kwargs.pop("sr"))
        frames = frame_array(x, 96, 48, pad=True, axis=0)
        _, embeddings = self.model(frames)
        return np.mean(embeddings.numpy(), 0)

    def process_batch(
        self, batch: Union[Iterable[np.ndarray], np.ndarray], **kwargs
    ) -> List[np.ndarray]:
        batch = list(batch)
        if all(is_mono_audio(x) for x in batch):
            sr = kwargs.pop("sr")
            batch = [_spectrogram(x, sr) for x in batch]
        frames = [frame_array(spec, 96, 48, pad=True, axis=0) for spec in batch]
        slices = np.cumsum([len(x) for x in frames])[:-1]
        frames = np.concatenate(frames)

        _, embeddings = self.model(frames)
        return [np.mean(x, 0) for x in np.split(embeddings.numpy(), slices)]

    def finish(self) -> None:
        import tensorflow as tf
        from keras.backend import clear_session

        clear_session()
        if self._old_env_log_level is not None:
            os.environ["TF_CPP_MIN_LOG_LEVEL"] = self._old_env_log_level
        tf.get_logger().setLevel(self._old_log_level)

    @property
    def dim(self) -> int:
        return 1024

    @property
    def is_sequence(self) -> bool:
        return False

    @property
    def feature_names(self) -> List[str]:
        return [f"yamnet_{i}" for i in range(self.dim)]


@dataclass
class VGGishExtractorConfig(ERTKConfig):
    """Configuration for a VGGish extractor."""

    model_dir: str = MISSING
    """The directory containing the VGGish model files."""
    postprocess: bool = True
    """Whether to apply the VGGish postprocessing."""


class VGGishExtractor(
    FeatureExtractor, AudioClipProcessor, fname="vggish", config=VGGishExtractorConfig
):
    """A VGGish extractor."""

    config: VGGishExtractorConfig

    def __init__(self, config: VGGishExtractorConfig) -> None:
        super().__init__(config)

        import tensorflow.compat.v1 as tf

        self._old_env_log_level = os.environ.get("TF_CPP_MIN_LOG_LEVEL")
        os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"  # Ignore INFO messages
        self._old_log_level = tf.get_logger().level
        tf.get_logger().setLevel(logging.WARNING)
        tf.disable_eager_execution()
        for gpu in tf.config.list_physical_devices("GPU"):
            tf.config.experimental.set_memory_growth(gpu, True)

        from ._vggish.vggish_postprocess import Postprocessor
        from ._vggish.vggish_slim import define_vggish_slim, load_vggish_slim_checkpoint

        pca_params = str(Path(config.model_dir, "vggish_pca_params.npz"))
        self.processor = Postprocessor(pca_params)

        self.sess = tf.Session()

        define_vggish_slim()
        ckpt = str(Path(config.model_dir, "vggish_model.ckpt"))
        load_vggish_slim_checkpoint(self.sess, ckpt)

        self.features_tensor = self.sess.graph.get_tensor_by_name(
            "vggish/input_features:0"
        )
        self.embedding_tensor = self.sess.graph.get_tensor_by_name("vggish/embedding:0")

    def finish(self) -> None:
        self.sess.close()
        if self._old_env_log_level is not None:
            os.environ["TF_CPP_MIN_LOG_LEVEL"] = self._old_env_log_level
        import tensorflow.compat.v1 as tf

        tf.get_logger().setLevel(self._old_log_level)

    def _process_frames(self, frames: np.ndarray) -> np.ndarray:
        [embeddings] = self.sess.run(
            [self.embedding_tensor], feed_dict={self.features_tensor: frames}
        )
        if self.config.postprocess:
            embeddings = self.processor.postprocess(embeddings).astype(np.float32)
        return embeddings

    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        if is_mono_audio(x):
            x = _spectrogram(x, kwargs.pop("sr"))
        frames = frame_array(x, 96, 96, pad=True, axis=0)
        return np.mean(self._process_frames(frames), 0)

    def process_batch(
        self, batch: Union[Iterable[np.ndarray], np.ndarray], **kwargs
    ) -> List[np.ndarray]:
        batch = list(batch)
        if all(is_mono_audio(x) for x in batch):
            sr = kwargs.pop("sr")
            batch = [_spectrogram(x, sr) for x in batch]
        frames = [frame_array(spec, 96, 96, pad=True, axis=0) for spec in batch]
        slices = np.cumsum([len(x) for x in frames])[:-1]
        frames = np.concatenate(frames)
        return [np.mean(x, 0) for x in np.split(self._process_frames(frames), slices)]

    @property
    def dim(self) -> int:
        return 128

    @property
    def is_sequence(self) -> bool:
        return False

    @property
    def feature_names(self) -> List[str]:
        return [f"vggish_{i}" for i in range(self.dim)]



"""Spectrogram extraction."""

import warnings
from dataclasses import dataclass
from typing import List, Optional

import librosa
import numpy as np

from ertk.config import ERTKConfig

from ._base import AudioClipProcessor, FeatureExtractor

__all__ = ["SpectrogramExtractorConfig", "SpectrogramExtractor", "spectrogram"]


def spectrogram(
    audio: np.ndarray,
    sr: float,
    kind: str = "mel",
    pre_emphasis: float = 0,
    window_size: float = 0.025,
    window_shift: float = 0.01,
    win_length_samp: Optional[int] = None,
    hop_length_samp: Optional[int] = None,
    n_fft: int = 2048,
    n_mels: int = 128,
    htk_mel: bool = False,
    n_chroma: int = 12,
    clip_db: Optional[float] = None,
    fmin: float = 0,
    fmax: Optional[float] = 8000,
    power: int = 2,
    to_log: Optional[str] = "db",
    mel_norm: str = "slaney",
):
    """General purpose spectrogram pipeline. Calculates spectrogram with
    optional pre-emphasis, clipping, mel/chroma transform, power and
    conversion to dB.

    Parameters
    ----------
    audio: np.ndarray
        Audio data. This should be a 1D array, or a 2D array with shape
        (1, n_samples).
    sr: int
        Sample rate.
    kind: str
        The kind of spectrogram to calculate. "mel" calculates a mel
        spectrogram, whereas "stft" just uses linear frequency.
    pre_emphasis: float
        Amount of pre-emphasis to apply. Default is 0 (no pre-emphasis).
    window_size: float
        Size of window in seconds. Default is 0.025 (25 ms).
    window_shift: float
        Amount window moves each frame, in seconds. Default is 0.01 (10
        ms).
    win_length_samp: int, optional
        The window length in samples. This overrides `window_size` if
        given.
    hop_length_samp: int, optional
        The hop size in samples. This overrides `window_shift` if given.
    n_fft: int, optional
        Number of FFT bins. Default is to use the next power of 2
        greater than `window_size * sr`.
    n_mels: int
        Number of mel bands.
    htk_mel: bool,
        Use HTK formula for mel calculation.
    n_chroma: int
        Number of chroma bands.
    clip_db: float, optional
        Whether to clip noise floor at a given level in dB below
        maximum. Default is `None` which does no clipping. This is only
        used if `to_db` is `True`.
    fmin: float
        Minimum frequency for mel bands. Default is 0.
    fmax: float, optional
        Maximum frequency for mel bands. Default is 8000.
    power: int
        Raise spectrogram magnitudes to this power. Default is 2 which
        is the usual power spectrogram.
    to_log: str
        Whether to convert spectrogram to logarithmic domain. Default is
        `db` which converts to dB units. If `to_log=='log'` then the
        natual logarithm is taken. Note that this argument is mainly
        only useful if `power == 2`.
    mel_norm: str
        Normalisation to apply to mel filters. Default is "slaney" as in
        librosa.

    Returns
    -------
    melspec: np.ndarray
        An array of shape (n_frames, n_mels) containing mel spectrogram.
        If power=2 this is the power spectrogram, and if to_db is `True`
        then this is in dB units.
    """
    audio = np.squeeze(audio)
    if audio.ndim != 1:
        raise ValueError(f"Audio should be in mono format, got shape {audio.shape}")

    win_length = win_length_samp if win_length_samp else int(window_size * sr)
    hop_length = hop_length_samp if hop_length_samp else int(window_shift * sr)

    # Pre-emphasis
    if pre_emphasis > 0:
        audio = librosa.effects.preemphasis(audio, pre_emphasis)

    # Spectrogram
    warnings.simplefilter("ignore", UserWarning)
    spec = librosa.stft(
        audio,
        n_fft=n_fft,
        hop_length=hop_length,
        win_length=win_length,
        center=True,
        window="hann",
        pad_mode="reflect",
    )
    spec = np.abs(spec) ** power
    if kind == "mel":
        try:
            mel_norm = int(mel_norm)
        except ValueError:
            pass
        spec = librosa.feature.melspectrogram(
            S=spec,
            n_mels=n_mels,
            fmin=fmin,
            fmax=fmax,
            htk=htk_mel,
            norm=mel_norm,
        )
    elif kind == "chroma":
        spec = librosa.feature.chroma_stft(S=spec, n_chroma=n_chroma)
    warnings.simplefilter("default", UserWarning)

    # Optional dB calculation
    if to_log == "db":
        spec = librosa.power_to_db(spec, ref=np.max, top_db=clip_db)
    elif to_log == "log":
        spec = np.log(np.maximum(spec, 1e-8))

    return spec.T


@dataclass
class SpectrogramExtractorConfig(ERTKConfig):
    """Configuration for SpectrogramExtractor."""

    kind: str = "mel"
    """The kind of spectrogram to calculate. "mel" calculates a mel
    spectrogram, whereas "stft" just uses linear frequency.
    """
    pre_emphasis: float = 0
    """Amount of pre-emphasis to apply. Default is 0 (no pre-emphasis)."""
    window_size: float = 0.025
    """Size of window in seconds. Default is 0.025 (25 ms)."""
    window_shift: float = 0.01
    """Amount window moves each frame, in seconds. Default is 0.01 (10 ms)."""
    win_length_samp: Optional[int] = None
    """The window length in samples. This overrides `window_size` if
    given.
    """
    hop_length_samp: Optional[int] = None
    """The hop size in samples. This overrides `window_shift` if given."""
    n_mels: int = 128
    """Number of mel bands."""
    htk_mel: bool = False
    """Use HTK formula for mel calculation."""
    n_fft: int = 2048
    """Number of FFT bins. Default is to use the next power of 2
    greater than `window_size * sr`.
    """
    n_chroma: int = 12
    """Number of chroma bands."""
    clip_db: Optional[float] = None
    """Whether to clip noise floor at a given level in dB below
    maximum. Default is `None` which does no clipping. This is only
    used if `to_db` is `True`.
    """
    fmin: float = 0
    """Minimum frequency for mel bands. Default is 0."""
    fmax: Optional[float] = 8000
    """Maximum frequency for mel bands. Default is 8000."""
    power: int = 2
    """Raise spectrogram magnitudes to this power. Default is 2 which
    is the usual power spectrogram.
    """
    to_log: Optional[str] = "db"
    """Whether to convert spectrogram to logarithmic domain. Default is
    `db` which converts to dB units. If `to_log=='log'` then the
    natual logarithm is taken. Note that this argument is mainly
    only useful if `power == 2`.
    """
    mel_norm: str = "slaney"
    """Normalisation to apply to mel filters. Default is "slaney" as in
    librosa.
    """


class SpectrogramExtractor(
    FeatureExtractor,
    AudioClipProcessor,
    fname="spectrogram",
    config=SpectrogramExtractorConfig,
):
    """Extracts a spectrogram from an audio clip."""

    config: SpectrogramExtractorConfig

    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        config = SpectrogramExtractorConfig.to_dictconfig(self.config)
        return spectrogram(x, sr=kwargs.pop("sr"), **config)

    @property
    def dim(self) -> int:
        if self.config.kind == "mel":
            return self.config.n_mels
        elif self.config.kind == "chroma":
            return self.config.n_chroma
        else:
            return 1 + self.config.n_fft // 2

    @property
    def is_sequence(self) -> bool:
        return True

    @property
    def feature_names(self) -> List[str]:
        return [f"{self.config.kind}_{i}" for i in range(self.dim)]



from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from ertk.dataset.annotation import read_annotations, write_annotations

from .constants import test_data_dir


def test_read_annotations_str_1() -> None:
    annotations = read_annotations(test_data_dir / "annot2_str.csv", dtype=str)
    assert type(annotations["extra1"]) is str
    assert annotations["extra1"] == "1"
    assert annotations["extra2"] == "0"
    assert annotations["1002_DFA_DIS_XX"] == "1"
    assert annotations["1002_DFA_HAP_XX"] == "3"


def test_read_annotations_str_2() -> None:
    annotations = read_annotations(test_data_dir / "label.csv", dtype=str)
    assert type(annotations["extra1"]) is str
    assert annotations["extra1"] == "none"
    assert annotations["extra2"] == "sadness"
    assert annotations["1002_DFA_DIS_XX"] == "disgust"
    assert annotations["1002_DFA_HAP_XX"] == "happiness"


def test_read_annotations_float() -> None:
    annotations = read_annotations(test_data_dir / "annot1.csv", dtype=float)
    assert isinstance(annotations["extra1"], (float, np.floating))
    assert annotations["extra1"] == 0
    assert annotations["extra2"] == 1
    assert annotations["1001_DFA_HAP_XX"] == pytest.approx(3.8472069)
    assert annotations["1001_DFA_NEU_XX"] == pytest.approx(6.9999)


def test_write_annotations_str(tmp_path: Path) -> None:
    mapping = {"clip1": "x", "clip2": "y", "clip3": "z"}
    out_path = tmp_path / "annotaton.csv"
    write_annotations(mapping, "test_annot", out_path)
    df = pd.read_csv(out_path, header=0, index_col=0, dtype=str)
    assert list(df.index) == ["clip1", "clip2", "clip3"]
    assert df.columns[0] == "test_annot"
    assert df.loc["clip1", "test_annot"] == "x"


def test_write_annotations_float(tmp_path: Path) -> None:
    mapping = {"clip1": 1.5, "clip2": 2, "clip3": 0.9}
    out_path = tmp_path / "annotaton_float.csv"
    write_annotations(mapping, "test_annot2", out_path)
    df = pd.read_csv(out_path, header=0, index_col=0, dtype={0: str, 1: float})
    assert list(df.index) == ["clip1", "clip2", "clip3"]
    assert df.columns[0] == "test_annot2"
    assert df.loc["clip2", "test_annot2"] == 2
    assert df.loc["clip3", "test_annot2"] == pytest.approx(0.9)



"""Miscellaneous utilities."""

import multiprocessing
import os
from functools import partial
from typing import Iterable

import joblib
import tqdm

__all__ = ["TqdmParallel", "TqdmMultiprocessing"]


# Class adapted from user394430's answer here:
# https://stackoverflow.com/a/61900501/10044861
# Licensed under CC BY-SA 4.0
class TqdmParallel(joblib.Parallel):
    """Convenience class that acts identically to joblib.Parallel except
    it uses a tqdm progress bar.

    Parameters
    ----------
    total: int
        Total number of items in the iterable.
    desc: str
        Progress bar description.
    unit: str
        Progress bar unit.
    leave: bool
        Whether to leave the progress bar after completion.
    **kwargs: dict
        Other keyword args passed to joblib.Parallel.
    """

    def __init__(
        self,
        total: int = 1,
        desc: str = "",
        unit: str = "it",
        leave: bool = True,
        **kwargs,
    ):
        self.tqdm_args = {
            "total": total,
            "desc": desc,
            "unit": unit,
            "leave": leave,
            "disable": None,
        }
        kwargs["verbose"] = 0
        super().__init__(**kwargs)

    def __call__(self, iterable):
        with tqdm.tqdm(**self.tqdm_args) as self.pbar:
            return super().__call__(iterable)

    def print_progress(self):
        self.pbar.n = self.n_completed_tasks
        self.pbar.refresh()


class TqdmMultiprocessing:
    def __init__(
        self,
        total: int = 1,
        desc: str = "",
        unit: str = "it",
        leave: bool = True,
        n_jobs: int = 1,
        context: str = "forkserver",
        **kwargs,
    ) -> None:
        self.n_jobs = n_jobs
        self.context = context
        self.tqdm_args = {
            "total": total,
            "desc": desc,
            "unit": unit,
            "leave": leave,
            "disable": None,
            **kwargs,
        }

    def imap(self, func, iterable, *args, chunksize=1, **kwargs) -> Iterable:
        _f = partial(func, *args, **kwargs)
        if self.n_jobs == -1:
            cpu_count = os.cpu_count()
            if cpu_count is None:
                raise RuntimeError("n_jobs == -1 but cannot get CPU count.")
            self.n_jobs = cpu_count
        self.pbar = tqdm.tqdm(iterable, **self.tqdm_args)
        if self.n_jobs > 1:
            with multiprocessing.get_context(self.context).Pool(self.n_jobs) as pool:
                yield from pool.imap(_f, self.pbar, chunksize=chunksize)
        else:
            yield from map(_f, iter(self.pbar))
        self.pbar.close()



import logging
import warnings
from abc import ABC, abstractmethod
from itertools import chain, tee
from typing import ClassVar, Dict, Iterable, List, Optional, Type, Union, cast

import librosa
import numpy as np
from importlib_metadata import EntryPoint, entry_points
from omegaconf import OmegaConf

from ertk.config import ERTKConfig
from ertk.utils import PathOrStr, batch_iterable


class InstanceProcessor(ABC):
    """An instance processor.

    Parameters
    ----------
    config: ERTKConfig
        The configuration for this processor.
    """

    config: ERTKConfig
    """The configuration for this processor."""
    logger: logging.Logger
    """The logger for this processor."""

    _config_type: ClassVar[Type[ERTKConfig]]
    _friendly_name: ClassVar[str]
    _registry: ClassVar[Dict[str, Type["InstanceProcessor"]]] = {}
    _plugins: ClassVar[Dict[str, EntryPoint]] = {
        x.name: x for x in entry_points().select(group="ertk.processors")
    }

    def __init__(self, config: ERTKConfig) -> None:
        self.config = config
        cls_name = f"{type(self).__module__}.{type(self).__name__}"
        self.logger = logging.getLogger(cls_name)

    def __init_subclass__(
        cls, fname: Optional[str] = None, config: Optional[Type[ERTKConfig]] = None
    ) -> None:
        cls._registry = {}
        if fname and config:
            cls._friendly_name = fname
            cls._config_type = config
            for t in cls.mro()[1:-1]:
                t = cast(Type[InstanceProcessor], t)  # For MyPy
                if not hasattr(t, "_registry"):
                    continue
                if fname in t._registry:
                    prev_cls = t._registry[fname]
                    msg = f"Name {fname} already registered with class {prev_cls}."
                    if prev_cls is cls:
                        warnings.warn(msg)
                    else:
                        raise KeyError(msg)
                t._registry[fname] = cls

    def __repr__(self):
        return f"{type(self).__name__}({self.config})"

    def __str__(self):
        yaml = OmegaConf.to_yaml(self.config).strip()
        yaml = "\t" + yaml.replace("\n", "\n\t")
        return f"{type(self)._friendly_name}(\n{yaml}\n)"

    @classmethod
    def friendly_name(cls) -> str:
        """Get the friendly name for this processor."""
        return cls._friendly_name

    @classmethod
    def get_processor_class(cls, name: str) -> "Type[InstanceProcessor]":
        """Get the class for the named processor.

        Parameters
        ----------
        name: str
            The name of the processor.

        Returns
        -------
        processor: Type[InstanceProcessor]
            The processor class.
        """
        if name not in cls._registry:
            ep = cls._plugins.get(name)
            if ep is None:
                raise KeyError(f"Unknown processor: {name}")
            try:
                obj = ep.load()
            except ImportError as e:
                raise ImportError(f"Plugin {ep.name} could not be loaded.") from e
            if not issubclass(obj, InstanceProcessor):
                raise TypeError(
                    f"Plugin {ep.name} of type {obj} does not subclass "
                    "InstanceProcessor"
                )
            cls._registry[name] = obj
        return cls._registry[name]

    @classmethod
    def make_processor(cls, name: str, config: ERTKConfig) -> "InstanceProcessor":
        """Create an instance of the named processor.

        Parameters
        ----------
        name: str
            The name of the processor to create.
        config: ERTKConfig
            The configuration for the processor.

        Returns
        -------
        processor: InstanceProcessor
            The created processor.
        """
        return cls.get_processor_class(name)(config)

    @classmethod
    def get_config_type(cls) -> Type[ERTKConfig]:
        """Get the configuration type for this processor."""
        return cls._config_type

    @classmethod
    def valid_processors(cls) -> List[str]:
        """Get a list of all registered processor names.

        Returns
        -------
        names: list of str
            The names of all registered processors.
        """
        return sorted(cls._registry.keys() | cls._plugins.keys())

    @classmethod
    def get_default_config(cls) -> ERTKConfig:
        """Get the default configuration for this processor.

        Returns
        -------
        config: ERTKConfig
            The default configuration.
        """
        return OmegaConf.structured(cls._config_type)

    @abstractmethod
    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        """Process a single audio clip.

        Parameters
        ----------
        x: np.ndarray
            The audio data to process.

        Returns
        -------
        result: np.ndarray
            The processed instance.
        """
        raise NotImplementedError()

    def process_batch(
        self, batch: Union[Iterable[np.ndarray], np.ndarray], **kwargs
    ) -> List[np.ndarray]:
        """Process a batch of instances. By default this simply calls
        `process_instance()` on each instance in the batch.

        Parameters
        ----------
        batch: np.ndarray or iterable of arrays
            The batch of instances to process.

        Returns
        -------
        processed: list of np.ndarray
            A list of processed instances.
        """
        return [self.process_instance(x, **kwargs) for x in batch]

    def process_all(
        self, xs: Union[Iterable[np.ndarray], np.ndarray], batch_size: int, **kwargs
    ) -> Iterable[np.ndarray]:
        """Process all instances in batches.

        Parameters
        ----------
        xs: iterable of np.ndarray
            The data from instances to process.
        batch_size: int
            Batch size.

        Returns
        -------
        processed: iterable of np.ndarray
            A generator that yields each processed instance in order.
        """
        if batch_size == 1:
            for x in xs:
                yield self.process_instance(x, **kwargs)
        elif batch_size > 1:
            yield from chain.from_iterable(
                self.process_batch(batch, **kwargs)
                for batch in batch_iterable(xs, batch_size)
            )
        elif batch_size < 0:
            yield from self.process_batch(xs, **kwargs)
        else:
            raise ValueError("Batch size cannot be 0.")

    def finish(self) -> None:
        """Perform any cleanup necesasry (e.g. closing files, unloading
        models, etc.)
        """

    @property
    @abstractmethod
    def feature_names(self) -> List[str]:
        """The names of the features produced by this processor."""
        raise NotImplementedError()


class AudioClipProcessor(InstanceProcessor):
    """Processes raw audio data."""

    def process_file(self, path: PathOrStr, sr: Optional[float] = None) -> np.ndarray:
        """Process individual audio file.

        Parameters
        ----------
        path: pathlike or str
            The path to an audio file.
        sr: float, optional
            Target sample rate. If not given, then each audio file is
            loaded with its native sample rate. Otherwise audio data is
            resampled to this sample rate.

        Returns
        -------
        processed: np.ndarray
            Processed instance.
        """
        audio, _sr = librosa.load(path, sr=sr, mono=True)
        return self.process_instance(audio, sr=_sr)

    def process_files(
        self, paths: Iterable[PathOrStr], batch_size: int, sr: Optional[float] = None
    ) -> Iterable[np.ndarray]:
        """Process a set of files.

        Parameters
        ----------
        paths: iterables of path or iterable of str
            Paths to audio files.
        batch_size: int
            Batch size. Note that if `batch_size != 1` then either `sr`
            must be given, or all files must have the same native sample
            rate.
        sr: float, optional
            Target sample rate. If not given, then each audio file is
            loaded with its native sample rate. Otherwise audio data is
            resampled to this sample rate.

        Returns
        -------
        processed: iterable of np.ndarray
            Iterable of processed instances.
        """
        if batch_size == 1:
            # Special case required for variable sample rate
            for path in paths:
                yield self.process_file(path, sr=sr)
        else:
            paths, tmp = tee(paths)
            _sr = librosa.load(next(tmp), sr=sr, duration=0)[1]
            audios = (librosa.load(path, sr=_sr, mono=True)[0] for path in paths)
            yield from self.process_all(audios, batch_size=batch_size, sr=_sr)


class FeatureExtractor(InstanceProcessor):
    """Extracts features from instances."""

    @property
    def dim(self) -> int:
        """The dimensionality of the extracted features."""
        return len(self.feature_names)

    @property
    @abstractmethod
    def is_sequence(self) -> bool:
        """Whether this FeatureExtractor yields sequence features."""
        raise NotImplementedError()



"""Process the raw BAVED dataset.

This assumes the file structure from the original data:
/.../
    0/
        *.wav
    1/
        *.wav
    ...
    speakers_info.json
"""

from pathlib import Path

import click

from ertk.dataset import resample_audio, write_annotations, write_filelist

emotion_level = ["low", "normal", "high"]
word_map = [
    "اعجبني",
    "لم يعجبني",
    "هذا",
    "الفيلم",
    "رائع",
    "مقول",
    "سيئ",
]


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the BAVED dataset at location INPUT_DIR."""
    paths = list(input_dir.glob("?/*.wav"))
    resample_dir = Path("resampled")
    if resample:
        resample_audio(paths, resample_dir)
    write_filelist(resample_dir.glob("*.wav"), "files_all")

    levels = {}
    speakers = {}
    genders = {}
    ages = {}
    words = {}
    for path in paths:
        spk, gen, age, word, level, _ = path.stem.split("-")
        levels[path.stem] = emotion_level[int(level)]
        speakers[path.stem] = spk
        genders[path.stem] = gen.upper()
        ages[path.stem] = int(age)
        words[path.stem] = word_map[int(word)]

    write_annotations(levels, "level")
    write_annotations(speakers, "speaker")
    write_annotations(genders, "gender")
    write_annotations(ages, "age")
    write_annotations(words, "word")
    write_annotations({p.stem: "ar" for p in paths}, "language")


if __name__ == "__main__":
    main()



"""Process the raw ShEMO dataset.

This assumes the file structure from the original compressed file:
/.../
    male/
        *.wav
    female/
    transcript/
"""

from pathlib import Path

import click

from ertk.dataset import resample_audio, write_annotations, write_filelist

emotion_map = {
    "A": "anger",
    "H": "happiness",
    "N": "neutral",
    "S": "sadness",
    "W": "surprise",
    "F": "fear",
}

unused_emotions = ["F"]


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the ShEMO dataset at location INPUT_DIR and resample
    audio to 16 kHz 16-bit WAV audio.
    """

    paths = list(input_dir.glob("*/*.wav"))
    if resample:
        resample_dir = Path("resampled")
        resample_audio(paths, resample_dir)
        write_filelist(resample_dir.glob("*.wav"), "files_all")
        write_filelist(
            [p for p in resample_dir.glob("*.wav") if p.stem[3] not in unused_emotions],
            "files_5class",
        )
    name2path = {p.stem: p for p in paths}

    write_annotations({n: emotion_map[n[3]] for n in name2path}, "label")
    speaker_dict = {n: n[:3] for n in name2path}
    write_annotations(speaker_dict, "speaker")
    write_annotations({k: v[0] for k, v in speaker_dict.items()}, "gender")
    write_annotations({n: "fa" for n in name2path}, "language")
    write_annotations({n: "ir" for n in name2path}, "country")

    transcripts = {}
    for trn in (input_dir / "transcript/final text").glob("*.ort"):
        with open(trn, encoding="utf_8_sig") as fid:
            transcripts[trn.stem] = (
                fid.read().strip().replace("\u200e", "").replace("\u200f", "")
            )
    write_annotations(transcripts, "transcript")
    missing = set(name2path) - transcripts.keys()
    write_filelist([name2path[x] for x in missing], "missing_transcripts")


if __name__ == "__main__":
    main()



"""Dataset features backend."""

import csv
import itertools
import logging
import typing
import warnings
from abc import ABC, abstractmethod
from collections import Counter
from pathlib import Path
from typing import (
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sized,
    Type,
    Union,
)

import arff
import librosa
import netCDF4
import numpy as np
import pandas as pd
import soundfile
from joblib import delayed

from ertk.dataset.utils import get_audio_paths, write_filelist
from ertk.utils import (
    PathOrStr,
    TqdmParallel,
    filter_kwargs,
    flat_to_inst,
    inst_to_flat,
    make_array_array,
)

__all__ = [
    "read_features",
    "read_features_iterable",
    "write_features",
    "register_format",
    "FeaturesData",
    "SequentialReader",
    "RandomAccessReader",
    "SequentialWriter",
    "NetCDFReader",
    "RawAudioReader",
    "ARFFSequentialReader",
    "CSVSequentialReader",
]

logger = logging.getLogger(__name__)


class FeaturesData:
    """Dataset features backend.

    Parameters
    ----------
    features: ndarray
        Features matrix. May be "flat" or per-instance. If flat,
        `slices` must also be present.
    names: list of str
        List of instance names.
    corpus: str
        Corpus for this dataset.
    feature_names: list of str, optional
        List of feature names. If not given, the features will me named
        feature1, feature2, etc.
    slices: list or ndarray:
        Slices corresponding to "flat" `features` matrix.
    dataset: DatasetBackend
        Other backend instance to copy attributes from. Takes priority
        over other passed args.
    """

    def __init__(
        self,
        features: Union[List[np.ndarray], np.ndarray],
        names: List[str],
        corpus: str = "",
        feature_names: Optional[List[str]] = None,
        slices: Union[np.ndarray, List[int], None] = None,
    ):
        self._corpus = corpus
        self._names = list(names)
        if isinstance(features, list):
            features = make_array_array(features)
        self._features = features
        if feature_names is None:
            feature_names = []
        self._feature_names = list(feature_names)

        if slices is not None:
            self._slices = np.array(slices)
            self._flat = features
            self._features = flat_to_inst(self.flat, slices)
        else:
            self.make_contiguous()

        n_features = self.features[0].shape[-1]
        if len(self.feature_names) == 0:
            self._feature_names = [f"feature{i + 1}" for i in range(n_features)]
        elif len(self.feature_names) != n_features:
            raise ValueError(
                f"feature_names has {len(self.feature_names)} entries but there "
                f"are {n_features} features"
            )

    def make_contiguous(self):
        """Makes `self.features` a contiguous array that represents
        views of `self.flat`.
        """
        self._flat, self._slices = inst_to_flat(self.features)
        self._features = flat_to_inst(self.flat, self.slices)

    def write(self, path: PathOrStr, **kwargs) -> None:
        """Write features to any file format."""

        write_features(
            path,
            self.features,
            names=self.names,
            corpus=self.corpus,
            feature_names=self.feature_names,
            **kwargs,
        )

    @property
    def features(self) -> np.ndarray:
        """Instance features matrix. This may be 2D/3D contiguous or
        variable length 3D.
        """
        return self._features

    @property
    def flat(self) -> np.ndarray:
        """2D "flat" features matrix. This has shape (N, F), where
        N = sum(slices), F is the number of features.
        """
        return self._flat

    @property
    def slices(self) -> np.ndarray:
        """Slices corresponding to flat features matrix. It should be
        the case that `features = flat_to_inst(flat, slices)` and
        `flat, slices = inst_to_flat(features)`.
        """
        return self._slices

    @property
    def names(self) -> List[str]:
        """Instance names."""
        return self._names

    @property
    def feature_names(self) -> List[str]:
        """Names of features in feature matrix."""
        return self._feature_names

    @property
    def corpus(self) -> str:
        """Corpus ID."""
        return self._corpus

    def __len__(self) -> int:
        return len(self.names)

    def __getitem__(self, index) -> np.ndarray:
        return self.features[index]


def read_arff_mem(path: PathOrStr, label: bool = False) -> FeaturesData:
    path = Path(path)
    with open(path) as fid:
        data = arff.load(fid)

    attr_names = [x[0] for x in data["attributes"]]
    counts = Counter([x[0] for x in data["data"]])
    idx = slice(1, -1) if label else slice(1, None)

    return FeaturesData(
        corpus=data["relation"],
        names=list(counts.keys()),
        features=np.array([x[idx] for x in data["data"]]),
        slices=np.array(list(counts.values())),
        feature_names=attr_names[idx],
    )


def read_csv_mem(
    path: PathOrStr, header: bool = True, label: bool = False
) -> FeaturesData:
    df = pd.read_csv(path, header=0 if header else None, converters={0: str})
    counts: typing.Counter[str] = Counter(df.iloc[:, 0])
    idx = slice(1, -1) if label else slice(1, None)

    return FeaturesData(
        names=list(counts.keys()),
        features=np.array(df.iloc[:, idx]),
        slices=np.array(list(counts.values())),
        feature_names=list(df.columns[idx]),
    )


def read_netcdf_mem(path: PathOrStr) -> FeaturesData:
    reader = NetCDFReader(path)
    features = reader[:]
    reader.close()
    return FeaturesData(features, reader.names, reader.corpus, reader.feature_names)


def _read_audio_file(path, sample_rate: int = 16000):
    warnings.filterwarnings("ignore", message="PySoundFile", category=UserWarning)
    audio, _ = librosa.load(path, sr=sample_rate, mono=True, res_type="kaiser_fast")
    return np.expand_dims(audio, -1)


def read_raw_mem(path: PathOrStr, sample_rate: int = 16000) -> FeaturesData:
    logger.info(f"Reading raw audio with sample_rate: {sample_rate}")
    path = Path(path)
    filepaths = get_audio_paths(path)
    _audio = TqdmParallel(len(filepaths), desc="Reading audio", leave=False, n_jobs=-1)(
        delayed(_read_audio_file)(filepath, sample_rate) for filepath in filepaths
    )

    return FeaturesData(
        corpus=path.parent.stem,
        names=[x.stem for x in filepaths],
        features=np.concatenate(_audio),
        slices=[len(x) for x in _audio],
        feature_names=["pcm"],
    )


class SequentialReader(Iterable[np.ndarray], Sized, ABC):
    """Base class for readers that read features sequentially."""

    feature_names: List[str]
    names: List[str]
    corpus: str = ""

    def __init__(self, path: PathOrStr, **kwargs) -> None:
        self.path = path
        self.open()

    def __len__(self):
        return len(self.names)

    def __iter__(self) -> Iterator[np.ndarray]:
        yield from self.read()

    @abstractmethod
    def open(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def read(self) -> Iterable[np.ndarray]:
        raise NotImplementedError()


class RandomAccessReader(SequentialReader, ABC):
    """Base class for readers that can access features by index."""

    @abstractmethod
    def __getitem__(
        self, index: Union[int, slice, List[int], np.ndarray]
    ) -> np.ndarray:
        raise NotImplementedError()

    def close(self) -> None:
        pass

    def read(self) -> Iterator[np.ndarray]:
        for i in range(len(self)):
            yield self[i]


class NetCDFReader(RandomAccessReader):
    """Read features from NetCDF4 file."""

    _is_2d: bool = False
    _is_3d: bool = False
    _is_vlen: bool = False

    def open(self) -> None:
        dataset = netCDF4.Dataset(self.path)
        self.corpus = getattr(dataset, "corpus", "")
        self.names = list(dataset.variables["name"])
        self.slices = np.asarray(dataset.variables["slices"])
        data_version = getattr(dataset, "ertk_data_version", "1")
        if data_version == "1":
            self.feature_names = list(dataset.variables["feature_names"])
            self._feat_var = "features"
        else:
            self.feature_names = list(dataset.variables["feature"])
            self._feat_var = "data"
        unique = np.unique(self.slices)
        if len(unique) == 1:
            if unique[0] == 1:
                self._is_2d = True
            else:
                self._is_3d = True
        else:
            self._is_vlen = True
        self.index = np.r_[0, np.cumsum(self.slices)]
        self.dataset = dataset

    def close(self) -> None:
        self.dataset.close()

    def __getitem__(
        self, index: Union[int, slice, List[int], np.ndarray]
    ) -> np.ndarray:
        feats_var = self.dataset.variables[self._feat_var]
        if self._is_2d:
            # Shortcut when we don't have to manipulate indices
            return np.asarray(feats_var[index])
        if isinstance(index, slice):
            start, stop, step = index.indices(len(self))
            if step != 1:
                raise ValueError("Non-contiguous slice (step != 1).")
            if start >= stop:
                raise ValueError("Empty slice.")
            istart = self.index[start]
            iend = self.index[stop]
            arr = np.asarray(feats_var[istart:iend])
            return flat_to_inst(arr, self.slices[start:stop])
        if isinstance(index, (list, np.ndarray)):
            _idx = np.array(index, dtype=np.int64)
            if _idx.ndim != 1:
                raise TypeError("If index is list/array, must be 1D.")
            items = [self.__getitem__(x) for x in _idx]
            if self._is_3d:
                return np.stack(items)
            else:
                return make_array_array(items)
        if isinstance(index, int):
            start, end = self.index[index], self.index[index + 1]
            return np.asarray(feats_var[start:end])
        raise NotImplementedError()


class RawAudioReader(RandomAccessReader):
    """Read raw audio files."""

    def __init__(self, path: PathOrStr, sample_rate: int = 16000, **kwargs) -> None:
        super().__init__(path, **kwargs)
        self.sample_rate = sample_rate

    def open(self) -> None:
        self.filepaths = np.array(get_audio_paths(self.path), dtype=object)
        self.names = [x.stem for x in self.filepaths]
        self.feature_names = ["pcm"]

    def __getitem__(
        self, index: Union[int, slice, List[int], np.ndarray]
    ) -> np.ndarray:
        if isinstance(index, int):
            return _read_audio_file(self.filepaths[index], self.sample_rate)
        return make_array_array(
            [
                _read_audio_file(filepath, self.sample_rate)
                for filepath in self.filepaths[index]
            ]
        )


class CSVSequentialReader(SequentialReader):
    """Read features from CSV file."""

    def open(self) -> None:
        with open(self.path) as fid:
            reader = csv.reader(fid)
            _, *feature_names = next(reader)
            self.feature_names = feature_names
            _names = [row[0] for row in reader]
            counts: typing.Counter[str] = Counter(_names)
            self.names = list(counts.keys())
            self.slices = np.array(list(counts.values()))

    def read(self) -> Iterator[np.ndarray]:
        with open(self.path) as fid:
            reader = csv.reader(fid)
            next(reader)  # Skip header
            for length in self.slices:
                yield np.array(
                    [x[1:] for x in itertools.islice(reader, length)]
                ).squeeze(0)


class ARFFSequentialReader(SequentialReader):
    """Read features from ARFF file."""

    def open(self) -> None:
        self.feature_names = []
        with open(self.path) as fid:
            for i, line in enumerate(fid):
                line = line.strip()
                if len(line) == 0:
                    continue
                if line.lower().startswith("@relation"):
                    self.corpus = line[10:].strip()
                elif line.lower().startswith("@attribute"):
                    feat_name, typ = line[11:].strip().rsplit(maxsplit=1)
                    if feat_name.lower() == "name":
                        continue
                    self.feature_names.append(feat_name)
                    if typ.lower() not in ["real", "numeric", "integer"]:
                        raise ValueError(f"Incorrect attribute type for {feat_name}.")
                elif line.lower().startswith("@data"):
                    break
            self._header = fid.tell()
            reader = csv.reader(fid, skipinitialspace=True)
            _names = [row[0] for row in reader]
            counts: typing.Counter[str] = Counter(_names)
            self.names = list(counts.keys())
            self.slices = np.array(list(counts.values()))

    def read(self) -> Iterator[np.ndarray]:
        self.feature_names = []
        with open(self.path) as fid:
            fid.seek(self._header)
            reader = csv.reader(fid, skipinitialspace=True)
            # rows = itertools.islice(reader, self._header, None)  # Skip header
            for length in self.slices:
                yield np.array([x[1:] for x in itertools.islice(reader, length)])


class SequentialWriter:
    """Write features to file.

    Parameters
    ----------
    names : List[str]
        List of utterance names.
    corpus : str
        Corpus name.
    feature_names : List[str]
        List of feature names.
    """

    def __init__(
        self,
        names: List[str],
        corpus: str = "",
        feature_names: List[str] = [],
    ) -> None:
        self.names = names
        self.corpus = corpus
        if len(feature_names) == 0:
            raise ValueError("Feature names must be provided.")
        self.feature_names = feature_names

    def write(self, path: PathOrStr, features: Iterable[np.ndarray], **kwargs):
        path = Path(path)
        path.parent.mkdir(exist_ok=True, parents=True)
        meth = SequentialWriter._write_backends.get(path.suffix)
        if meth is None:
            raise ValueError(f"File format {path.suffix} not supported.")
        meth(self, path, features, **kwargs)

    def write_arff(self, path: PathOrStr, features: Iterable[np.ndarray], **kwargs):
        """Write features to ARFF file."""

        with open(path, "w") as fid:
            fid.write(f"@RELATION {self.corpus}\n\n")
            fid.write("@ATTRIBUTE name STRING\n")
            for feat in self.feature_names:
                fid.write(f"@ATTRIBUTE {feat} NUMERIC\n")
            fid.write("\n@data\n")
            writer = csv.writer(fid)
            for name, inst in zip(self.names, features):
                if len(inst.shape) == 2:
                    for row in inst:
                        writer.writerow([name] + list(row))
                else:
                    writer.writerow([name] + list(inst))

    def write_csv(self, path, features: Iterable[np.ndarray], header=True, **kwargs):
        """Write features to CSV file."""

        with open(path, "w") as fid:
            writer = csv.writer(fid, lineterminator="\n")
            if header:
                writer.writerow(["name"] + self.feature_names)
            for name, inst in zip(self.names, features):
                if len(inst.shape) == 2:
                    for row in inst:
                        writer.writerow([name] + list(row))
                else:
                    writer.writerow([name] + list(inst))

    def write_netcdf(
        self,
        path: PathOrStr,
        features: Iterable[np.ndarray],
        chunksize: int = 1024,
        **kwargs,
    ):
        """Write features to NetCDF4 file."""

        dataset = netCDF4.Dataset(path, "w")
        dataset.createDimension("name", len(self.names))
        dataset.createDimension("concat", 0)

        # Need to get first array for metadata
        features, _tmp = itertools.tee(features)
        _x0 = next(_tmp)
        del _tmp
        dim = _x0.shape[-1]
        dtype = _x0.dtype
        if dtype.kind == "f":
            # Force single-precision for consistency and space
            dtype = np.float32
        dataset.createDimension("feature", dim)

        filename = dataset.createVariable("name", str, ("name",))
        filename[:] = np.array(self.names)

        _features = dataset.createVariable(
            "data", dtype, ("concat", "feature"), chunksizes=(chunksize, dim)
        )
        slices = []
        idx = 0
        for x in features:
            i = 1 if x.ndim == 1 else len(x)
            _features[idx : idx + i, :] = x
            slices.append(i)
            idx += i

        _slices = dataset.createVariable("slices", int, ("name",))
        _slices[:] = np.array(slices)

        _feature_names = dataset.createVariable("feature", str, ("feature",))
        _feature_names[:] = np.array(self.feature_names)

        dataset.setncattr_string("corpus", self.corpus)
        dataset.setncattr_string("ertk_data_version", "2")
        dataset.close()

    def write_raw(
        self, path: PathOrStr, features: Iterable[np.ndarray], sr: int = 16000, **kwargs
    ):
        """Write raw audio files."""

        path = Path(path)
        output_dir = path.with_suffix("")
        output_dir.mkdir(exist_ok=True, parents=True)
        for name, audio in zip(self.names, features):
            output_path = output_dir / f"{name}.wav"
            soundfile.write(output_path, audio, subtype="PCM_16", samplerate=sr)
        write_filelist(output_dir.glob("*.wav"), path.with_suffix(".txt"))

    _write_backends: Dict[str, Callable] = {
        ".arff": write_arff,
        ".csv": write_csv,
        ".nc": write_netcdf,
        ".txt": write_raw,
        "": write_raw,
    }


_READ_BACKENDS_MEM: Dict[str, Callable[..., FeaturesData]] = {
    ".nc": read_netcdf_mem,
    ".arff": read_arff_mem,
    ".csv": read_csv_mem,
    ".txt": read_raw_mem,
}

_READ_BACKENDS_SEQUENTIAL: Dict[str, Type[SequentialReader]] = {
    ".nc": NetCDFReader,
    ".arff": ARFFSequentialReader,
    ".csv": CSVSequentialReader,
    ".txt": RawAudioReader,
}


def register_format(
    suffix: str,
    read: Optional[Callable[..., FeaturesData]],
    write: Optional[Callable[..., None]],
):
    """Register new read/write functions for a given file format."""

    if suffix[0] != ".":
        suffix = "." + suffix
    if read is not None:
        _READ_BACKENDS_MEM[suffix] = read
    if write is not None:
        SequentialWriter._write_backends[suffix] = write


def find_features_file(files: Iterable[Path]) -> Path:
    files = list(files)
    for suffix in _READ_BACKENDS_MEM:
        file = next((x for x in files if x.suffix == suffix), None)
        if file is not None:
            return file
    raise FileNotFoundError(f"No features found in {files}")


def read_features(path: PathOrStr, **kwargs) -> FeaturesData:
    """Read features from given path into memory."""

    path = Path(path)
    meth = _READ_BACKENDS_MEM.get(path.suffix)
    if meth is None:
        raise ValueError(f"File format {path.suffix} not supported.")
    logger.debug(f"Reading features from {path}")
    return meth(path, **filter_kwargs(kwargs, meth))


def read_features_iterable(path: PathOrStr, **kwargs) -> SequentialReader:
    """Read features from given path either sequentially or by index."""

    path = Path(path)
    meth = _READ_BACKENDS_SEQUENTIAL.get(path.suffix)
    if meth is None:
        raise ValueError(f"File format {path.suffix} not supported.")
    logger.debug(f"Reading features from {path}")
    return meth(path, **filter_kwargs(kwargs, meth.read))


def write_features(
    path: PathOrStr,
    features: Union[Iterable[np.ndarray], np.ndarray],
    names: List[str],
    corpus: str = "",
    feature_names: List[str] = [],
    slices: Union[np.ndarray, List[int], None] = None,
    **kwargs,
) -> None:
    """Convenience function to write features to given path.

    Parameters
    ----------
    path : PathOrStr
        Path to write features to.
    features : Union[Iterable[np.ndarray], np.ndarray]
        Features to write. If a 3D array is given, the dimensions are
        assumed to be (instances, frames, features). If a 2D array is
        given, the dimensions are assumed to be (frames, features), and
        the `slices` parameter is required. If an iterable of arrays is
        given, each 1D or 2D array is assumed to be a single instance.
    names : List[str]
        Instance names.
    corpus : str
        Name of corpus. Default is empty string.
    feature_names : List[str]
        Feature names.
    slices : Union[np.ndarray, List[int]], optional
        Number of frames per instance. Only required if a 2D matrix is
        given, and the first dimension is different than the number of
        instances.
    """
    if slices is not None:
        if not isinstance(features, np.ndarray):
            raise ValueError("`slices` should only be given for a 2D contiguous array.")
        features = flat_to_inst(features, slices)

    SequentialWriter(names=names, corpus=corpus, feature_names=feature_names).write(
        path, features, **kwargs
    )



"""Process the raw EmoFilm dataset.

This assumes the file structure from the original compressed file:
/.../
    *.wav
    f_m_corpus_it_es_en_legend.xlsx
"""

from pathlib import Path

import click
import pandas as pd

from ertk.dataset import resample_audio, write_annotations, write_filelist

emotion_map = {
    "ans": "fear",
    "dis": "disgust",
    "gio": "happiness",
    "rab": "anger",
    "tri": "sadness",
}

film_map = {
    "a bautiful mind": "a beautiful mind",
    "beautiful mind": "a beautiful mind",
    "american hisotry x": "american history x",
    "million dolar baby": "million dollar baby",
    "joe black": "meet joe black",
    "bicentennian man": "bicentennial man",
    "ther interpreter": "the interpreter",
    "poltergueist": "poltergeist",
    "will hinting": "good will hunting",
    "borkeback mountain": "brokeback mountain",
    "brokeback": "brokeback mountain",
    "borkebackmountain": "brokeback mountain",
    "the live of the others": "the lives of others",
    "the life of the others": "the lives of others",
}

speaker_map = {
    "camilo garcia": "camilo garcía",
    "constantino romero+6": "constantino romero",
    "adalberto maria merli+6": "adalberto maria merli",
    "giuppi izzo": "giuppy izzo",
    "maria moscardo": "maría moscardó",
    "massimo de ambrosio": "massimo de ambrosis",
    "robert duval": "robert duvall",
    "xavier fernandez": "xavier fernández",
}


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the EmoFilm dataset at location INPUT_DIR and resample
    audio to 16 kHz 16-bit WAV audio.
    """
    paths = list(input_dir.glob("*.wav"))
    if resample:
        resample_dir = Path("resampled")
        resample_audio(paths, resample_dir)
        write_filelist(resample_dir.glob("*.wav"), "files_all")

    write_annotations({p.stem: emotion_map[p.stem[2:5]] for p in paths}, "label")

    df = pd.concat(
        pd.read_excel(input_dir / "f_m_corpus_it_es_en_legend.xlsx", sheet_name=None)
    ).set_index("file")
    df["gender"] = df.index.map(lambda x: x[0])
    df["language"] = df.index.map(lambda x: x[-2:])
    df["film"] = df["film"].str.lower().replace(film_map)
    df["speaker"] = (
        df["speaker"]
        .str.strip()
        .str.lower()
        .replace(r"^\?+$", pd.NA, regex=True)
        .str.replace("\u00a0", " ", regex=False)
        .replace(speaker_map)
    )
    for col in df.columns:
        write_annotations(df[col].to_dict(), col)


if __name__ == "__main__":
    main()



"""Process the raw EMOVO dataset.

This assumes the file structure from the original compressed file:
/.../
    f1/
        *.wav
    m1/
        *.wav
    ...
"""

from pathlib import Path

import click

from ertk.dataset import resample_audio, write_annotations, write_filelist

emotion_map = {
    "dis": "disgust",
    "gio": "happiness",
    "neu": "neutral",
    "pau": "fear",
    "rab": "anger",
    "sor": "surprise",
    "tri": "sadness",
}

sentence_map = {
    "b1": "Gli operai si alzano presto",
    "b2": "I vigili sono muniti di pistola",
    "b3": "La cascata fa molto rumore",
    "l1": "L'autunno prossimo Tony partirà per la Spagna nella prima metà di ottobre",
    "l2": "Ora prendo la felpa di là ed esco per fare una passeggiata",
    "l3": "Un attimo dopo s'è incamminato ... ed è inciampato",
    "l4": "Vorrei il numero telefonico del Signor Piatti",
    "n1": "La casa forte vuole col pane",
    "n2": "La forza trova il passo e l'aglio rosso",
    "n3": "Il gatto sta scorrendo nella pera",
    "n4": "Insalata pastasciutta coscia d'agnello limoncello",
    "n5": "Uno quarantatré dieci mille cinquantasette venti",
    "d1": "Sabato sera cosa farà?",
    "d2": "Porti con te quella cosa?",
}


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the EMOVO dataset at location INPUT_DIR and resample
    audio to 16 kHz 16-bit WAV audio.
    """
    paths = list(input_dir.glob("**/*.wav"))
    resample_dir = Path("resampled")
    if resample:
        resample_audio(paths, resample_dir)
    write_filelist(resample_dir.glob("*.wav"), "files_all")

    write_annotations({p.stem: emotion_map[p.stem[0:3]] for p in paths}, "label")
    write_annotations({p.stem: p.stem[4:6] for p in paths}, "speaker")
    write_annotations({p.stem: p.stem[4].upper() for p in paths}, "gender")
    write_annotations({p.stem: p.stem[7:9] for p in paths}, "sentence")
    write_annotations({p.stem: sentence_map[p.stem[7:9]] for p in paths}, "transcript")
    write_annotations({p.stem: "it" for p in paths}, "language")
    write_annotations({p.stem: "it" for p in paths}, "country")


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import Tuple, Union

import click
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from click_option_group import RequiredAnyOptionGroup, optgroup
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import MinMaxScaler, Normalizer, StandardScaler

from ertk.dataset import DataLoadConfig, Dataset, load_datasets_config, read_features


@click.group()
def main():
    """Tools for visualising data."""


@main.command("xy")
@optgroup.group("input", cls=RequiredAnyOptionGroup)
@optgroup.option(
    "--conf", "data_conf", type=click.Path(exists=True, dir_okay=False, path_type=Path)
)
@optgroup.option(
    "--features", type=click.Path(exists=True, dir_okay=False, path_type=Path)
)
@optgroup.option(
    "--corpus",
    "corpus_info",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
)
@click.option(
    "--transform",
    type=click.Choice(["pca", "umap", "tsne", "std", "minmax"]),
    default=("pca",),
    multiple=True,
)
@click.option(
    "--colour",
    "colour_part",
    type=str,
    help="Colour by this partition.",
)
@click.option(
    "--style",
    "style_part",
    type=str,
    help="Style markers by this partition.",
)
@click.option(
    "--sample",
    "sample_str",
    type=str,
    default="1.0",
    help="Proportion or number of instances to sample for visualisation",
)
@click.argument("restargs", type=str, nargs=-1)
def xy(
    data_conf: Path,
    features: Path,
    corpus_info: Path,
    transform: Tuple[str],
    colour_part: str,
    style_part: str,
    sample_str: str,
    restargs: Tuple[str],
):
    """Plots INPUT in 2D using one of a number of transforms. INPUT can
    either be a features file or data load config.
    """

    if data_conf or corpus_info:
        if data_conf:
            print(f"Loading data from {data_conf}")
            conf = DataLoadConfig()
            conf = DataLoadConfig.from_file(data_conf, override=list(restargs))
            data = load_datasets_config(conf)
        else:
            if features is None:
                raise ValueError("Must specify --features to use with --corpus")
            print(f"Loading data from {features}")
            data = Dataset(corpus_info, features=features)

        x = data.x
        hue = data.get_annotations(colour_part) if colour_part else None
        style = data.get_annotations(style_part) if style_part else None
    else:
        print(f"Loading data from {features}")
        x = read_features(features).features
        hue = None
        style = None

    try:
        sample = int(float(sample_str) * len(x))
    except ValueError:
        sample = int(sample_str)
    print(f"Using {sample} instances for visualisation")
    if sample < len(x):
        rng = np.random.default_rng()
        perm = rng.permutation(len(x))[:sample]
        x = x[perm]
        if hue is not None:
            hue = hue[perm]

    trn = []
    for trans in transform:
        if trans == "pca":
            trn.append(PCA(n_components=2))
        elif trans == "tsne":
            trn.append(TSNE(n_components=2))
        elif trans == "minmax":
            trn.append(MinMaxScaler())
        elif trans == "std":
            trn.append(StandardScaler())
        elif trans == "norm":
            trn.append(Normalizer())
        elif trans == "umap":
            from umap import UMAP  # Import here because of import time

            trn.append(UMAP(n_components=2))
    pipeline = make_pipeline(*trn)

    print(f"Transforming with {pipeline}")
    x_new = pipeline.fit_transform(x)
    plt.figure()
    sns.scatterplot(x=x_new[:, 0], y=x_new[:, 1], hue=hue, style=style)
    plt.show()


@main.command("feats")
@click.argument("input", type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.argument("instance", type=str, default="2")
def feats(input: Path, instance: str):
    """Displays plot of INSTANCE in INPUT. INSTANCE can either be a
    numeric index, a range of indices using numpy slice notation or a
    named instance.
    """

    data = read_features(input)
    if instance.isdigit():
        idx: Union[int, slice] = int(instance)
    else:
        _i = instance.find(":")
        if _i != -1:
            start = int(instance[:_i])
            end = int(instance[_i + 1 :])
            idx = slice(start, end)
        else:
            idx = data.names.index(instance)
    arr = data.features[idx]
    names = data.names[idx]
    print(names)

    plt.figure()
    plt.imshow(arr, aspect="equal", origin="upper", interpolation="nearest")
    plt.xlabel("Features")
    plt.ylabel("Instance" if len(names) > 1 else "Time")
    plt.show()


if __name__ == "__main__":
    main()



"""Multi-task learning MTL utilities."""

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Callable, Dict, List

import omegaconf
import torch

from ertk.config import ERTKConfig
from ertk.pytorch.utils import get_loss

from ._base import ERTKPyTorchModel, PyTorchModelConfig

__all__ = ["MTLModelConfig", "MTLModel", "MTLTaskConfig"]


@dataclass
class MTLTaskConfig(ERTKConfig):
    output_dim: int = omegaconf.MISSING
    loss: str = omegaconf.MISSING
    metrics: List[str] = field(default_factory=list)
    weight: float = 1


@dataclass
class MTLModelConfig(PyTorchModelConfig):
    tasks: Dict[str, MTLTaskConfig] = omegaconf.MISSING


class MTLModel(ERTKPyTorchModel, ABC):
    tasks: Dict[str, MTLTaskConfig]
    losses: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
    config: MTLModelConfig

    def __init__(self, config: MTLModelConfig) -> None:
        super().__init__(config)
        self.tasks = config.tasks
        self.losses = {n: get_loss(t.loss) for n, t in self.tasks.items()}

    @abstractmethod
    def forward(self, x: torch.Tensor, **kwargs) -> Dict[str, torch.Tensor]:
        raise NotImplementedError()

    def training_step(self, batch, batch_idx: int) -> torch.Tensor:
        x, ys = batch
        outputs = self.forward(x)
        total_loss = 0
        accs = {k: 0 for k in self.tasks}
        losses = {k: 0 for k in self.tasks}
        for k, v in outputs.items():
            loss = self.losses[k](v, ys[k])
            losses[k] = loss
            total_loss = total_loss + loss * self.tasks[k].weight
            y_pred = v.argmax(1)
            accs[k] = torch.sum(y_pred == ys[k]).item() / float(len(ys[k]))
        res = {
            "loss/train": total_loss,
            **{f"acc_{k}/train": accs[k] for k in accs},
            **{f"loss_{k}/train": losses[k] for k in losses},
        }
        self.log_dict(res)
        return total_loss

    def validation_step(self, batch, batch_idx: int) -> None:
        x, ys = batch
        outputs = self.forward(x)
        total_loss = 0
        accs = {k: 0 for k in self.tasks}
        losses = {k: 0 for k in self.tasks}
        for k, v in outputs.items():
            # print(k, ys[k])
            loss = self.losses[k](v, ys[k])
            losses[k] += loss
            total_loss = total_loss + loss * self.tasks[k].weight
            y_pred = v.argmax(1)
            accs[k] = torch.sum(y_pred == ys[k]).item() / float(len(ys[k]))
        self.log_dict(
            {
                "loss/valid": total_loss,
                **{f"acc_{k}/valid": accs[k] for k in accs},
                **{f"loss_{k}/valid": losses[k] for k in losses},
            }
        )

    def predict_step(self, batch, batch_idx: int) -> torch.Tensor:
        return self.forward(batch[0])

    def configure_optimizers(self):
        return torch.optim.Adam(self.parameters(), lr=self.config.learning_rate)



from typing import Tuple

import numpy as np
import pytest

from ertk.utils.array import (
    batch_arrays_by_length,
    check_3d,
    clip_arrays,
    flat_to_inst,
    frame_array,
    frame_arrays,
    inst_to_flat,
    make_array_array,
    pad_array,
    pad_arrays,
    shuffle_multiple,
    transpose_time,
)


def test_make_array_array_vlen():
    arr = make_array_array([np.arange(i) for i in [3, 5, 7, 11]])
    assert arr.dtype == object
    assert arr.shape == (4,)
    assert [len(x) for x in arr] == [3, 5, 7, 11]


def test_make_array_array_3d():
    arr = make_array_array(list(np.arange(1000).reshape(10, 10, 10)))
    assert arr.shape == (10,)
    assert all(x.shape == (10, 10) for x in arr)


def test_check_3d():
    check_3d(np.arange(1000).reshape(10, 10, 10))
    check_3d([np.arange(3 * i).reshape(3, i) for i in [3, 5, 7, 11]])
    with pytest.raises(ValueError):
        check_3d(np.arange(100).reshape(10, 10))


@pytest.mark.parametrize(
    ["shape", "frame_size", "frame_shift", "pad", "axis", "expected_shape"],
    [
        ((100,), 30, 10, False, 0, (8, 30)),
        ((107,), 30, 10, True, 0, (9, 30)),
        ((100, 10), 30, 10, False, 0, (8, 30, 10)),
        ((100, 10), 30, 10, True, 0, (8, 30, 10)),
        ((107, 10), 30, 10, False, 0, (8, 30, 10)),
        ((107, 10), 30, 10, True, 0, (9, 30, 10)),
        ((110, 10), 30, 10, False, 0, (9, 30, 10)),
        ((47, 10), 13, 5, False, 0, (7, 13, 10)),
        ((47, 10), 13, 5, True, 0, (8, 13, 10)),
        ((20, 100, 10), 30, 10, False, 1, (20, 8, 30, 10)),
        ((20, 100, 10), 30, 10, True, 1, (20, 8, 30, 10)),
        ((1, 37, 10), 50, 50, True, 1, (1, 1, 50, 10)),
        ((37, 10), 50, 50, True, 0, (1, 50, 10)),
        ((37, 10), 50, 10, True, 0, (1, 50, 10)),
        ((49, 64), 96, 48, True, 0, (1, 96, 64)),
        ((48, 64), 96, 48, True, 0, (1, 96, 64)),
    ],
)
def test_frame_array_shape(
    shape: Tuple[int],
    frame_size: int,
    frame_shift: int,
    pad: bool,
    axis: int,
    expected_shape: Tuple[int],
):
    rng = np.random.default_rng()
    x = rng.random(size=shape)
    framed = frame_array(x, frame_size, frame_shift, pad=pad, axis=axis, copy=True)
    assert framed.shape == expected_shape
    if not pad:
        framed = frame_array(x, frame_size, frame_shift, pad=pad, axis=axis, copy=False)
        assert framed.shape == expected_shape


@pytest.mark.parametrize(
    ["shape", "frame_size", "frame_shift", "axis"],
    [
        ((1, 37, 10), 50, 50, 1),
        ((37, 10), 50, 50, 0),
        ((37, 10), 50, 10, 0),
    ],
)
def test_frame_array_small_nopad(
    shape: Tuple[int], frame_size: int, frame_shift: int, axis: int
):
    rng = np.random.default_rng()
    x = rng.random(size=shape)
    with pytest.raises(ValueError, match="sequence is shorter than frame_size"):
        frame_array(x, frame_size, frame_shift, pad=False, axis=axis)


@pytest.mark.parametrize(
    ["shape", "frame_size", "frame_shift", "axis"],
    [
        ((107,), 30, 10, 0),
        ((47, 10), 13, 5, 0),
        ((47, 10), 13, 5, 0),
    ],
)
def test_frame_array_pad_nocopy(
    shape: Tuple[int], frame_size: int, frame_shift: int, axis: int
):
    rng = np.random.default_rng()
    x = rng.random(size=shape)
    with pytest.raises(ValueError, match="pad must be False"):
        frame_array(x, frame_size, frame_shift, pad=True, axis=axis, copy=False)


def test_frame_array_content_no_pad():
    rng = np.random.default_rng()
    x = rng.random(size=(105, 10))
    framed = frame_array(x, 30, 10, pad=False, axis=0)
    assert np.array_equal(framed[0], x[0:30])
    assert np.array_equal(framed[1], x[10:40])
    assert np.array_equal(framed[7], x[70:100])


def test_frame_array_content_pad():
    rng = np.random.default_rng()
    x = rng.random(size=(107, 10))
    framed = frame_array(x, 30, 10, pad=True, axis=0)
    assert np.array_equal(framed[0], x[0:30])
    assert np.array_equal(framed[1], x[10:40])
    assert np.array_equal(framed[7], x[70:100])
    assert np.array_equal(framed[8, 0:27], x[80:107])
    assert np.all(framed[8, 27:30] == 0)


def test_frame_arrays_cont():
    arrays = [np.arange(i).reshape(-1, 1) for i in [100, 207, 150]]
    framed = frame_arrays(arrays, 30, 17)
    assert framed.shape == (3, 11, 30, 1)
    assert np.array_equal(framed[0, 0, :, 0], np.arange(30))
    assert np.array_equal(framed[1, 10, :, 0], np.arange(170, 200))


def test_frame_arrays_vlen():
    arrays = [np.arange(i).reshape(-1, 1) for i in [100, 207, 150]]
    framed = frame_arrays(arrays, 30, 17, vlen=True)
    assert framed.shape == (3,)
    assert np.array_equal(framed[0][0, :, 0], np.arange(30))
    assert np.array_equal(framed[1][10, :, 0], np.arange(170, 200))


def test_frame_arrays_2d():
    arrays = [np.arange(i) for i in [100, 207, 150]]
    with pytest.raises(ValueError):
        frame_arrays(arrays, 30, 10)


@pytest.mark.parametrize(
    ["shape", "to_multiple", "to_size", "axis", "expected_shape"],
    [
        ((107, 10), 10, None, 0, (110, 10)),
        ((107, 10), 5, None, 0, (110, 10)),
        ((107, 10), 20, None, 0, (120, 10)),
        ((55, 10), None, 60, 0, (60, 10)),
        ((30, 10), None, 60, 0, (60, 10)),
        ((5, 30, 10), None, 60, 1, (5, 60, 10)),
        ((5, 30, 10), None, 20, 2, (5, 30, 20)),
        ((5, 30, 10), 50, None, 1, (5, 50, 10)),
    ],
)
def test_pad_array(
    shape: Tuple[int],
    to_multiple: int,
    to_size: int,
    axis: int,
    expected_shape: Tuple[int],
):
    rng = np.random.default_rng()
    x = rng.random(size=shape)
    padded = pad_array(x, to_multiple, to_size, axis=axis)
    assert padded.shape == expected_shape
    idx_same = tuple(slice(0, x) for x in shape)
    assert np.array_equal(padded[idx_same], x)
    idx_zeros = [slice(None, None) for _ in shape]
    idx_zeros[axis] = slice(shape[axis], None)
    assert np.all(padded[tuple(idx_zeros)] == 0)


def test_pad_arrays_type():
    arrays = [np.arange(i) for i in [100, 207, 150]]
    padded = pad_arrays(arrays, 32)
    assert isinstance(padded, list)
    arrays = make_array_array([np.arange(i) for i in [100, 207, 150]])
    padded = pad_arrays(arrays, 32)
    assert isinstance(padded, np.ndarray)


def test_pad_arrays_vlen():
    arrays = [np.arange(i) for i in [100, 207, 150]]
    padded = pad_arrays(arrays, 32)
    assert len(padded) == 3
    assert padded[0].shape == (128,)


def test_pad_arrays_3d():
    x = np.arange(1000).reshape(10, 10, 10)
    padded = pad_arrays(x, 32)
    assert padded.shape == (10, 32, 10)


def test_pad_arrays_cont():
    arrays = make_array_array([np.arange(i) for i in [10, 11, 12]])
    padded = pad_arrays(arrays, 32)
    assert padded.shape == (3, 32)


def test_clip_arrays_type():
    arrays = [np.arange(i) for i in [100, 207, 150]]
    clipped = clip_arrays(arrays, 130)
    assert isinstance(clipped, list)
    arrays = make_array_array([np.arange(i) for i in [100, 207, 150]])
    clipped = clip_arrays(arrays, 130)
    assert isinstance(clipped, np.ndarray)


def test_clip_arrays_vlen():
    arrays = [np.arange(i) for i in [100, 207, 150]]
    clipped = clip_arrays(arrays, 130)
    assert clipped[0].shape == (100,)
    assert clipped[1].shape == (130,)
    assert clipped[2].shape == (130,)


def test_clip_arrays_3d():
    arrays = np.arange(1000).reshape(10, 10, 10)
    clipped = clip_arrays(arrays, 7)
    assert clipped.shape == (10, 7, 10)


def test_clip_arrays_cont():
    arrays = make_array_array([np.arange(i) for i in [100, 207, 150]])
    clipped = clip_arrays(arrays, 80)
    assert clipped.shape == (3, 80)


def test_transpose_time_exception():
    with pytest.raises(ValueError):
        transpose_time([np.arange(i) for i in [100, 207, 150]])


def test_transpose_time_3d():
    rng = np.random.default_rng()
    x = rng.random(size=(10, 100, 13))
    transposed = transpose_time(x.copy())
    assert np.array_equal(transposed, x.transpose(0, 2, 1))


def test_transpose_time_vlen():
    arrays = [np.arange(i * 3).reshape(i, 3) for i in [100, 207, 150]]
    transposed = transpose_time(arrays)
    assert transposed[0].shape == (3, 100)
    assert transposed[1].shape == (3, 207)
    assert np.array_equal(transposed[2], np.arange(150 * 3).reshape(150, 3).T)


def test_shuffle_multiple():
    array1 = [0, 1, 6, 7]
    array2 = [12, 17, 14, 13]
    sh1, sh2 = shuffle_multiple(array1, array2, numpy_indexing=False)
    assert set(sh1) == set(array1)
    assert set(sh2) == set(array2)


def test_shuffle_multiple_exception():
    array1 = [0, 1, 6, 7]
    array2 = [12, 17, 14]
    with pytest.raises(ValueError):
        shuffle_multiple(array1, array2, numpy_indexing=False)


def test_batch_arrays_by_length():
    arrays = [np.arange(i) for i in [20, 5, 5, 20]]
    batch_x, batch_y = batch_arrays_by_length(
        arrays, np.arange(4), batch_size=2, shuffle=False
    )
    assert batch_x.shape == (2,)
    assert batch_x[0].shape == (2, 5)
    assert batch_x[1].shape == (2, 20)
    assert batch_y.shape == (2,)
    assert list(batch_y[0]) == [1, 2]
    assert list(batch_y[1]) == [0, 3]

    batch_x, batch_y = batch_arrays_by_length(
        arrays, np.arange(4), batch_size=2, shuffle=False, uniform_batch_size=True
    )
    assert batch_y.shape == (2, 2)


def test_batch_arrays_by_length_nonuniform():
    arrays = [np.arange(i) for i in [20, 5, 5, 22]]
    batch_x, batch_y = batch_arrays_by_length(
        arrays, np.arange(4), batch_size=2, shuffle=False
    )
    assert batch_x.shape == (3,)
    assert batch_x[0].shape == (2, 5)
    assert batch_x[1].shape == (1, 20)
    assert batch_x[2].shape == (1, 22)
    assert batch_y.shape == (3,)
    assert list(batch_y[0]) == [1, 2]
    assert list(batch_y[1]) == [0]
    assert list(batch_y[2]) == [3]


def test_flat_to_inst_2d():
    flat = np.arange(1000).reshape(100, 10)
    inst = flat_to_inst(flat, [1] * 100)
    assert inst is flat


def test_flat_to_inst_3d():
    flat = np.arange(1000).reshape(100, 10)
    inst = flat_to_inst(flat, [20] * 5)
    assert np.array_equal(inst, flat.reshape(5, 20, 10))


def test_flat_to_inst_vlen():
    flat = np.arange(1000).reshape(100, 10)
    inst = flat_to_inst(flat, [30, 10, 10, 40, 10])
    assert inst.shape == (5,)
    assert np.array_equal(inst[0], flat[0:30])
    assert np.array_equal(inst[1], flat[30:40])
    assert np.array_equal(inst[4], flat[90:100])


def test_inst_to_flat_2d():
    inst = np.arange(1000).reshape(100, 10)
    flat, slices = inst_to_flat(inst)
    assert flat is inst
    assert list(slices) == [1] * 100


def test_inst_to_flat_3d():
    inst = np.arange(1000).reshape(5, 20, 10)
    flat, slices = inst_to_flat(inst)
    assert np.array_equal(flat, inst.reshape(100, 10))
    assert list(slices) == [20] * 5


def test_inst_to_flat_vlen():
    inst = make_array_array([np.arange(i).reshape(-1, 1) for i in [20, 5, 5, 22]])
    flat, slices = inst_to_flat(inst)
    assert flat.shape == (52, 1)
    assert list(slices) == [20, 5, 5, 22]
    assert np.array_equal(flat, np.concatenate(inst))



"""
OpenXBOW extractor
==================

This module contains the OpenXBOW extractor, which is a wrapper around
the OpenXBOW Java program.

.. autosummary::
    :toctree:

    OpenXBOWExtractor
    OpenXBOWExtractorConfig
"""

from .openxbow import OpenXBOWExtractor, OpenXBOWExtractorConfig



"""PyTorch classification utilities."""

import logging
import warnings
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import pytorch_lightning as pl
import torch
import torch.nn as nn
from pytorch_lightning.callbacks import Callback, ModelCheckpoint, ModelSummary
from pytorch_lightning.loggers import CSVLogger, Logger, TensorBoardLogger
from sklearn.model_selection import BaseCrossValidator, LeaveOneGroupOut
from torch.utils.data import DataLoader, TensorDataset

from ertk.pytorch.dataset import MTLDataModule
from ertk.pytorch.models import LightningWrapper
from ertk.pytorch.models.mtl import MTLModel
from ertk.pytorch.train import PyTorchTrainConfig
from ertk.train import ExperimentResult, get_scores
from ertk.utils import ScoreFunction

__all__ = [
    "pt_cross_validate",
    "pt_train_val_test",
    "train_mtl_model",
]


def pt_cross_validate(
    model_fn: Callable[..., Union[nn.Module, pl.LightningModule]],
    x: np.ndarray,
    y: np.ndarray,
    *,
    groups: Optional[np.ndarray] = None,
    cv: BaseCrossValidator = LeaveOneGroupOut(),
    scoring: Union[
        str, List[str], Dict[str, ScoreFunction], Callable[..., float]
    ] = "accuracy",
    verbose: int = 0,
    n_jobs: int = 1,
    fit_params: Dict[str, Any] = {},
) -> ExperimentResult:
    sw = fit_params.pop("sample_weight", None)

    logging.debug(f"cv={cv}")
    logging.debug(f"sample_weight={sw}")

    n_folds = cv.get_n_splits(x, y, groups)
    scores = defaultdict(list)
    preds = np.zeros((len(y), len(np.unique(y))), dtype=np.float32)
    for fold, (train, test) in enumerate(cv.split(x, y, groups)):
        if verbose:
            logging.info(f"Fold {fold + 1}/{n_folds}")

        x_train = x[train]
        y_train = y[train]
        x_test = x[test]
        y_test = y[test]
        sw_train = sw[train] if sw is not None else None
        sw_test = sw[test] if sw is not None else None

        result = pt_train_val_test(
            model_fn,
            train_data=(x_train, y_train, sw_train),
            valid_data=(x_test, y_test, sw_test),
            test_data=(x_test, y_test, sw_test),
            scoring=scoring,
            verbose=verbose,
            fit_params=dict(fit_params),
        )

        for k, v in result.scores_dict.items():
            scores[k].append(v)
        preds[test] = result.predictions
    return ExperimentResult({k: np.array(scores[k]) for k in scores}, predictions=preds)


def pt_train_val_test(
    model_fn: Callable[..., Union[nn.Module, pl.LightningModule]],
    train_data: Tuple[np.ndarray, ...],
    valid_data: Tuple[np.ndarray, ...],
    test_data: Optional[Tuple[np.ndarray, ...]] = None,
    scoring: Union[
        str, List[str], Dict[str, ScoreFunction], Callable[..., float]
    ] = "accuracy",
    verbose: int = 0,
    fit_params: Dict[str, Any] = {},
) -> ExperimentResult:
    train_config: PyTorchTrainConfig = fit_params.pop("train_config")
    log_dir = train_config.logging.log_dir
    batch_size = train_config.batch_size

    logging.debug(f"train_config={train_config}")

    loggers: List[Logger] = []
    callbacks: List[Callback] = []
    if verbose:
        callbacks.append(ModelSummary(5))
    if log_dir:
        _logger = TensorBoardLogger(log_dir, name=None)
        loggers.append(_logger)
        loggers.append(CSVLogger(_logger.log_dir, name=None, version=""))
        if train_config.enable_checkpointing:
            callbacks.append(
                ModelCheckpoint(
                    dirpath=_logger.log_dir,
                    save_last=True,
                    every_n_epochs=train_config.save_every_n_epochs,
                    filename="checkpoint-{epoch}_{step}",
                    verbose=verbose > 0,
                )
            )

    if test_data is None:
        test_data = valid_data

    clf = model_fn()
    logging.debug(clf)
    if isinstance(clf, tuple):  # 'pipeline' tuple
        transform = clf[0]
        clf = clf[1]
        transform.fit(train_data[0], y=train_data[1])
        train_data = (transform.transform(train_data[0]), *train_data[1:])
        valid_data = (transform.transform(valid_data[0]), *valid_data[1:])
        test_data = (transform.transform(test_data[0]), *test_data[1:])

    if not isinstance(clf, pl.LightningModule):
        clf = LightningWrapper(clf, train_config.wrapper_model_config)

    train_data_pt = tuple(torch.from_numpy(x) for x in train_data)
    valid_data_pt = tuple(torch.from_numpy(x) for x in valid_data)
    test_data_pt = tuple(torch.from_numpy(x) for x in test_data)

    train_ds = TensorDataset(*train_data_pt)
    valid_ds = TensorDataset(*valid_data_pt)
    test_ds = TensorDataset(*test_data_pt)
    train_dl = DataLoader(
        train_ds, batch_size=batch_size, shuffle=True, pin_memory=True
    )
    valid_dl = DataLoader(
        valid_ds, batch_size=batch_size, shuffle=False, pin_memory=True
    )
    test_dl = DataLoader(test_ds, batch_size=batch_size, shuffle=False, pin_memory=True)

    if verbose <= 0:
        logging.getLogger("pytorch_lightning").setLevel(logging.ERROR)
        warnings.simplefilter("ignore", UserWarning)
    trainer = pl.Trainer(
        gpus=train_config.n_gpus,
        strategy=train_config.dist_strategy,
        max_epochs=train_config.epochs,
        logger=loggers if loggers else False,
        enable_progress_bar=bool(verbose),
        enable_checkpointing=train_config.enable_checkpointing,
        enable_model_summary=False,
        callbacks=callbacks,
    )
    trainer.fit(
        clf,
        train_dataloaders=train_dl,
        val_dataloaders=valid_dl,
        ckpt_path=train_config.resume_checkpoint,
    )
    y_pred = torch.cat(trainer.predict(clf, test_dl)).cpu().numpy()
    if verbose <= 0:
        logging.getLogger("pytorch_lightning").setLevel(logging.INFO)
        warnings.simplefilter("default", UserWarning)

    y_true = test_data[1]
    scores = get_scores(scoring, np.argmax(y_pred, -1), y_true)
    scores = {f"test_{k}": v for k, v in scores.items()}
    return ExperimentResult(scores, predictions=y_pred)


def train_mtl_model(
    model: MTLModel,
    tasks: Dict[int, int],
    data: Optional[MTLDataModule],
    train_data: Optional[DataLoader] = None,
    valid_data: Optional[DataLoader] = None,
    epochs: int = 200,
    device: str = "cuda:0",
    verbose: int = 0,
):
    trainer = pl.Trainer(
        gpus=1,
        max_epochs=epochs,
        logger=bool(verbose),
        enable_progress_bar=bool(verbose),
        enable_model_summary=bool(verbose),
        enable_checkpointing=False,
    )
    cols = [f"{x}_acc" for x in tasks] + [f"{x}_loss" for x in tasks]
    stats_df = pd.DataFrame(
        columns=pd.Index(cols, name="metric"),
        index=pd.MultiIndex.from_product(
            [range(epochs), ["train", "val"]], names=["epoch", "phase"]
        ),
    )

    if data is not None:
        trainer.fit(model, data)
        [val] = trainer.validate(model, data, verbose=verbose)
    else:
        trainer.fit(model, train_dataloaders=train_data, val_dataloaders=valid_data)
        [val] = trainer.validate(model, valid_data, verbose=verbose)

    return val



from dataclasses import dataclass, field
from typing import Any, Dict

from ertk.config import ERTKConfig

__all__ = ["TFModelConfig"]


@dataclass
class TFModelConfig(ERTKConfig):
    """Configuration for a TensorFlow model."""

    optimiser: str = "adam"
    """The optimiser to use."""
    opt_params: Dict[str, Any] = field(default_factory=dict)
    """The parameters to pass to the optimiser."""
    learning_rate: float = 1e-3
    """The learning rate to use."""
    n_features: int = -1
    """The number of features in the input data."""
    n_classes: int = -1
    """The number of classes in the output data."""
    loss: str = "sparse_categorical_crossentropy"
    """The loss function to use."""



"""Process the raw RAVDESS dataset.

This assumes the file structure from the original compressed file:
/.../
    Audio/
        Actor_01/
            *.wav
        ...
    ...
"""

from pathlib import Path

import click

from ertk.dataset import resample_audio, write_annotations, write_filelist

emotion_map = {
    "01": "neutral",
    "02": "calm",
    "03": "happiness",
    "04": "sadness",
    "05": "anger",
    "06": "fear",
    "07": "disgust",
    "08": "surprise",
}

statement_map = {
    "01": "Kids are talking by the door",
    "02": "Dogs are sitting by the door",
}


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the RAVDESS dataset at location INPUT_DIR and resample
    audio to 16 kHz 16-bit WAV audio.
    """

    paths = list(input_dir.glob("Audio/Actor_??/03-*.wav"))
    if resample:
        resample_dir = Path("resampled")
        resample_audio(paths, resample_dir)
        paths = list(resample_dir.glob("*.wav"))
    write_filelist(paths, "files_all")
    write_filelist([x for x in paths if x.stem.startswith("03-01-")], "files_speech")
    write_filelist([x for x in paths if x.stem.startswith("03-02-")], "files_song")

    write_annotations(
        {p.stem: ["song", "speech"][int(p.stem[3:5]) % 2] for p in paths}, "style"
    )
    write_annotations({p.stem: emotion_map[p.stem[6:8]] for p in paths}, "label")
    speaker_dict = {p.stem: p.stem[-2:] for p in paths}
    write_annotations(speaker_dict, "speaker")
    write_annotations(
        {k: ["F", "M"][int(v) % 2] for k, v in speaker_dict.items()}, "gender"
    )
    write_annotations({p.stem: "en" for p in paths}, "language")
    write_annotations({p.stem: "ca" for p in paths}, "country")
    write_annotations({p.stem: p.stem[9:11] for p in paths}, "intensity")
    write_annotations({p.stem: p.stem[12:14] for p in paths}, "statement")
    write_annotations({p.stem: p.stem[15:17] for p in paths}, "repetition")
    write_annotations(
        {p.stem: statement_map[p.stem[12:14]] for p in paths}, "transcript"
    )


if __name__ == "__main__":
    main()



from pathlib import Path

from constants import (
    corpus_name,
    feature_names,
    features_2d,
    features_3d,
    features_vlen,
    slices_vlen,
)

from ertk.dataset import write_features


def main():
    base_dir = Path(__file__).parent / "test_data"
    assert base_dir.exists()

    with open(base_dir / "all_clips.txt") as fid:
        names = [Path(x.strip()).stem for x in fid]

    for format in [".arff", ".csv", ".nc"]:
        write_features(
            base_dir / "features" / ("features_2d" + format),
            features_2d,
            names,
            corpus=corpus_name,
            feature_names=feature_names,
        )
        write_features(
            base_dir / "features" / ("features_3d" + format),
            features_3d,
            names,
            corpus=corpus_name,
            feature_names=feature_names,
        )
        write_features(
            base_dir / "features" / ("features_vlen" + format),
            features_vlen,
            names,
            corpus=corpus_name,
            feature_names=feature_names,
            slices=slices_vlen,
        )


if __name__ == "__main__":
    main()



"""Process the raw JL dataset.

This assumes the file structure from the original compressed file:
/.../
    Raw JL Corpus (unchecked ...)/
        JL(wav+txt)/
            *.wav
        ...
    ...
"""

import re
from pathlib import Path

import click

from ertk.dataset import resample_audio, write_annotations, write_filelist

REGEX = re.compile(r"^(?:fe)?male[12]_([a-z]+)_\d+[ab]_[12]$")

emotion_map = {
    "angry": "anger",
    "sad": "sadness",
    "neutral": "neutral",
    "happy": "happiness",
    "excited": "excitedness",
    "anxious": "anxiety",
    "apologetic": "apologetic",
    "assertive": "assertive",
    "concerned": "concern",
    "encouraging": "encouraging",
}

secondary_emotions = ["anxious", "apologetic", "assertive", "concerned", "encouraging"]


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the JL dataset at location INPUT_DIR and resample audio
    to 16 kHz 16-bit WAV audio.
    """

    paths = list(
        input_dir.glob("Raw JL corpus (unchecked and unannotated)/JL(wav+txt)/*.wav")
    )

    transcripts = {}
    for p in paths:
        trn_file = p.with_suffix(".txt")
        with open(trn_file, "r") as fid:
            transcripts[p.stem] = fid.read().strip()
    write_annotations(transcripts, "transcript")

    if resample:
        resample_dir = Path("resampled")
        resample_audio(paths, resample_dir)
        paths = list(resample_dir.glob("*.wav"))
    write_filelist([p for p in paths], "files_all")
    write_filelist(
        [p for p in paths if REGEX.match(p.stem).group(1) not in secondary_emotions],
        "files_primary",
    )
    write_filelist(
        [p for p in paths if REGEX.match(p.stem).group(1) in secondary_emotions],
        "files_secondary",
    )

    write_annotations(
        {p.stem: emotion_map.get(REGEX.match(p.stem).group(1)) for p in paths},
        "label",
    )
    speaker_dict = {p.stem: p.stem[: p.stem.find("_")] for p in paths}
    write_annotations(speaker_dict, "speaker")
    write_annotations({k: v[0].upper() for k, v in speaker_dict.items()}, "gender")
    write_annotations({p.stem: "en" for p in paths}, "language")
    write_annotations({p.stem: "nz" for p in paths}, "country")
    write_annotations({p.stem: p.stem[-4:-2] for p in paths}, "sentence")


if __name__ == "__main__":
    main()



from pathlib import Path

import librosa
import numpy as np

from ertk.dataset.utils import (
    get_audio_paths,
    resample_audio,
    resample_rename_clips,
    write_filelist,
)

from .constants import audio_dir, resample_dir, resample_paths, test_data_dir


def test_get_audio_paths_dir() -> None:
    paths = get_audio_paths(audio_dir)
    assert len(paths) == 12


def test_get_audio_paths_file_1() -> None:
    paths = get_audio_paths(test_data_dir / "all_clips.txt")
    assert len(paths) == 12


def test_get_audio_paths_file_2() -> None:
    paths = get_audio_paths(test_data_dir / "subset.txt")
    assert len(paths) == 6


# This one is redundant since test_data_dir is absolute due to __file__
def test_get_audio_paths_absolute() -> None:
    paths = get_audio_paths(audio_dir, absolute=True)
    for path in paths:
        assert path.is_absolute()
        assert path == audio_dir / path.name
    assert len(paths) == 12


def test_resample(tmp_path: Path) -> None:
    resample_audio(resample_paths, tmp_path)
    paths = list(tmp_path.glob("*.wav"))
    for path in paths:
        audio, sr = librosa.load(path, sr=None)
        assert sr == 16000
        ref, _ = librosa.load(resample_dir / path.name, sr=None)
        assert np.allclose(audio, ref)
    assert len(paths) == 12


def test_resample_rename(tmp_path: Path) -> None:
    mapping = {x: tmp_path / (x.stem + "_rename.wav") for x in resample_paths}
    resample_rename_clips(mapping)
    paths = list(tmp_path.glob("*.wav"))
    for path in paths:
        assert path.name.endswith("_rename.wav")
        _, sr = librosa.load(path, sr=None, duration=0)
        assert sr == 16000
    assert len(paths) == 12


def test_write_filelist_path(tmp_path: Path) -> None:
    paths = ["test_clip1.wav", "clip2.mp3", "clip3.ogg"]
    out_path = tmp_path / "test_filelist.txt"
    write_filelist([Path(x) for x in paths], out_path)
    with open(out_path) as fid:
        lines = list(map(str.strip, fid))
    assert lines == sorted(paths)



"""OpenSMILE feature extraction."""

from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List

import numpy as np
from omegaconf import MISSING

from ertk.config import ERTKConfig
from ertk.utils import is_mono_audio

from ._base import AudioClipProcessor, FeatureExtractor

__all__ = ["OpenSMILEExtractorConfig", "OpenSMILEExtractor"]


@dataclass
class OpenSMILEExtractorConfig(ERTKConfig):
    """OpenSMILE feature extractor configuration."""

    opensmile_config: str = MISSING
    """OpenSMILE config file."""
    levels: List[str] = field(default_factory=lambda: ["func"])
    """OpenSMILE levels to extract."""
    opensmile_opts: Dict[str, str] = field(default_factory=dict)
    """Additional OpenSMILE options."""


class OpenSMILEExtractor(
    FeatureExtractor,
    AudioClipProcessor,
    fname="opensmile",
    config=OpenSMILEExtractorConfig,
):
    """OpenSMILE feature extractor."""

    config: OpenSMILEExtractorConfig

    def __init__(self, config: OpenSMILEExtractorConfig) -> None:
        import opensmile

        super().__init__(config)

        opensmile_config = config.opensmile_config
        if not Path(opensmile_config).exists():
            opensmile_config = opensmile.FeatureSet[opensmile_config]

        self.levels = config.levels
        self.smiles = [
            opensmile.Smile(
                opensmile_config, level, options=dict(config.opensmile_opts)
            )
            for level in config.levels
        ]
        self._feature_names = []
        for smile in self.smiles:
            self._feature_names.extend(smile.feature_names)

    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        sr = kwargs.pop("sr")
        x = np.squeeze(x)
        if not is_mono_audio(x):
            raise ValueError("Audio must be mono")
        xs = [smile(x, sr) for smile in self.smiles]
        # TODO: allow other window/padding correction options
        cut = min(x.shape[2] for x in xs)
        xs = [x[:, :, :cut] for x in xs]
        x = np.concatenate(xs, axis=1)
        # Get single segment and transpose axes so that frames are first
        return x[0].T

    @property
    def is_sequence(self) -> bool:
        return "func" not in self.levels

    @property
    def feature_names(self) -> List[str]:
        return self._feature_names



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, Iterable, List, Union

import numpy as np
from omegaconf import MISSING

from ertk.config import ERTKConfig
from ertk.utils import PathOrStr

from ._base import InstanceProcessor


@dataclass
class ProcessorConfig:
    """Configuration for a processor."""

    type: str = MISSING
    """The type of processor to use (friendly name)."""
    batch_size: int = 1
    """The batch size to use for processing."""
    config: Dict[str, Any] = field(default_factory=lambda: {})
    """The processor configuration."""


class ChainingType(Enum):
    """The type of chaining to use for a processing pipeline."""

    IN_MEMORY = "in_memory"
    """Results from each processor are stored in memory."""
    GENERATOR = "generator"
    """Results from each processor are yielded as they are produced."""


@dataclass
class ProcessingPipelineConfig(ERTKConfig):
    """Configuration for a processing pipeline."""

    input: str = MISSING
    """The input data to process."""
    output: str = MISSING
    """Where to write the output data."""
    chaining: ChainingType = ChainingType.IN_MEMORY
    """The type of chaining to use for the pipeline"""
    pipeline: List[str] = MISSING
    """The processors to use in the pipeline, in order"""
    processors: Dict[str, ProcessorConfig] = MISSING
    """The processors to use in the pipeline, and their configurations"""


class ProcessingPipeline(InstanceProcessor):
    """A processing pipeline.

    Parameters
    ----------
    config: ProcessingPipelineConfig
        The configuration for the pipeline.
    """

    pipeline: List[InstanceProcessor]
    """The processors in the pipeline."""
    chaining: ChainingType = ChainingType.IN_MEMORY
    """The type of chaining to use for the pipeline"""
    output: str
    """Where to write the output data."""
    config: ProcessingPipelineConfig
    """The pipeline configuration."""

    def __init__(self, config: ProcessingPipelineConfig) -> None:
        from ertk.dataset.features import read_features_iterable

        data = read_features_iterable(config.input)
        self.data = data
        self.config = config
        self.chaining = config.chaining
        self.output = config.output
        self.pipeline = []
        for name in config.pipeline:
            # This implicitly checks that the processor exists, and that
            # the config is valid
            proc_conf = (
                InstanceProcessor.get_processor_class(name)
                .get_config_type()
                .from_config(config.processors[name].config)
            )
            processor = InstanceProcessor.make_processor(name, proc_conf)
            self.pipeline.append(processor)

    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        for processor in self.pipeline:
            x = processor.process_instance(x, **kwargs)
        return x

    def process_batch(
        self, batch: Union[Iterable[np.ndarray], np.ndarray], **kwargs
    ) -> List[np.ndarray]:
        for processor in self.pipeline:
            batch = processor.process_batch(batch, **kwargs)
        return list(batch)

    def process_all(
        self, xs: Union[Iterable[np.ndarray], np.ndarray], batch_size: int, **kwargs
    ) -> Iterable[np.ndarray]:
        for processor in self.pipeline:
            xs = processor.process_all(xs, batch_size, **kwargs)
            if self.chaining == ChainingType.IN_MEMORY:
                xs = list(xs)
        return xs

    def finish(self) -> None:
        for processor in self.pipeline:
            processor.finish()

    @property
    def feature_names(self) -> List[str]:
        return self.pipeline[-1].feature_names


def load_pipeline(file: PathOrStr) -> ProcessingPipeline:
    """Load a processing pipeline from a YAML file.

    Parameters
    ----------
    file: pathlike or str
        The path to the YAML file.

    Returns
    -------
    pipeline: ProcessingPipeline
        The loaded processing pipeline.
    """

    config = ProcessingPipelineConfig.from_file(file)
    return ProcessingPipeline(config)



from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from ertk.cli.dataset.convert import main


def test_noargs():
    with pytest.raises(SystemExit) as excinfo:
        main([])
    assert excinfo.value.code == 2


def test_badargs():
    with pytest.raises(SystemExit) as excinfo:
        main(["x", "y"])
    assert excinfo.value.code == 2


test_data_dir = Path(__file__).parent / "../../test_data"


def test_convert(tmp_path):
    with pytest.raises(SystemExit) as excinfo:
        main(
            [
                f"{test_data_dir}/features/features_2d.nc",
                f"{tmp_path}/converted.csv",
            ]
        )
    assert excinfo.value.code == 0
    df_converted = pd.read_csv(f"{tmp_path}/converted.csv", index_col=0)
    df_original = pd.read_csv(f"{test_data_dir}/features/features_2d.csv", index_col=0)
    assert df_converted.columns.equals(df_original.columns)
    assert np.allclose(df_converted.values, df_original.values, equal_nan=True)



"""Process the raw VENEC dataset.

This assumes the file structure from the original compressed file:
/.../
    NoNameProsody/
        *.mp3
        *.mp3.wav [duplicated by mistake?]
    CowenEtAlAnalysisCode/
        ...
    veneccountryinfo.csv
    ...
"""

import re
from pathlib import Path

import click
import pandas as pd

from ertk.dataset import resample_audio, write_annotations, write_filelist

MED_INT_REGEX = re.compile(r"^(\d\d?)([A-Za-z]+)$")
HIGH_INT_REGEX = re.compile(r"^[A-Za-z]+(\d+)([A-Za-z]+)\d?(?:[Hh]igh|Hi|H)$")
EMOTIONS = {
    "aff": "affection",
    "amu": "amusement",
    "ang": "anger",
    "con": "contempt",
    "disg": "disgust",
    "dist": "distress",
    "fea": "fear",
    "gui": "guilt",
    "hap": "happiness",
    "int": "interest",
    "lus": "lust",
    "negs": "negativesurprise",
    "negsurprise": "negativesurprise",
    "neu": "neutral",
    "nsur": "negativesurprise",
    "nsu": "negativesurprise",
    "poss": "positivesurprise",
    "possurprise": "positivesurprise",
    "psur": "positivesurprise",
    "pri": "pride",
    "rel": "relief",
    "sad": "sadness",
    "ser": "serenity",
    "sha": "shame",
    "sham": "shame",
}


def get_speaker(s: str):
    try:
        return MED_INT_REGEX.match(s)[1]
    except TypeError:
        return HIGH_INT_REGEX.match(s)[1]


def get_emotion(s: str):
    try:
        emo = MED_INT_REGEX.match(s)[2].lower()
    except TypeError:
        emo = HIGH_INT_REGEX.match(s)[2].lower()
    return EMOTIONS.get(emo, emo)


NO_SIGNAL = {"1528", "0191", "2080", "0723"}
MISSING_LABEL = {"0731", "1370", "2370", "2505"}


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the VENEC dataset at location INPUT_DIR and resample
    audio to 16 kHz 16-bit WAV audio.
    """

    paths = [
        x
        for x in input_dir.glob("NoNameProsody/*.mp3")
        if x.stem not in NO_SIGNAL | MISSING_LABEL
    ]
    resample_dir = Path("resampled")
    if resample:
        resample_audio(paths, resample_dir)
    write_filelist(resample_dir.glob("*.wav"), "files_all")

    clip_info = pd.read_csv(
        input_dir / "veneccountryinfo.csv",
        index_col=0,
        header=None,
        names=["Filename", "Country", "Original"],
    )
    clip_info.index = clip_info.index.map(lambda x: x[:-4])

    # Acted labels
    write_annotations(
        clip_info["Original"].map(lambda x: Path(x).stem).map(get_emotion).to_dict(),
        "label",
    )
    write_annotations(
        clip_info["Original"].map(lambda x: Path(x).parts[0].split("_")[0]).to_dict(),
        "intensity",
    )

    csvfile = input_dir / "CategoryRatingsUSA.csv"
    ratings = pd.read_csv(csvfile, index_col=0, header=0)
    ratings.index = ratings.index.map(lambda x: x[:-4])

    mode_count = ratings.apply(lambda x: x.value_counts().sort_index().iloc[-1], axis=1)
    ratings["Freq"] = ratings.max(1)
    ratings["Majority"] = ratings.columns[ratings.to_numpy().argmax(1)]
    ratings["Num_Majority"] = mode_count == 1

    agreement = ratings["Freq"].mean()
    print(f"Mean label agreement: {agreement:.3f}")

    ratings = ratings.join(clip_info)
    ratings["Country"] = ratings["Country"].map(str.upper)
    # Correct the three errors where country is STR
    ratings.loc[ratings["Country"] == "STR", "Country"] = "AUS"
    # Intensity is High or Medium
    ratings["Intensity"] = ratings["Original"].map(
        lambda x: Path(x).parts[0].split("_")[0]
    )
    ratings["Speaker_ID"] = (
        ratings["Country"]
        + "_"
        + ratings["Original"]
        .map(lambda x: Path(x).stem)
        .map(get_speaker)
        .map(int)
        .map("{:02d}".format)
    )
    ratings["Emotion"] = (
        ratings["Original"].map(lambda x: Path(x).stem).map(get_emotion)
    )

    ratings_18class = ratings[ratings["Emotion"] != "neutral"]

    # (country, speaker) pairs for which there are clips for all
    # emotions. This gives around 2262 clips.
    ratings_complete = ratings_18class.groupby("Speaker_ID").filter(
        lambda x: set(x["Emotion"]) >= set(EMOTIONS.values()) - {"neutral"}
    )

    write_filelist(
        [p for p in resample_dir.glob("*.wav") if p.stem in ratings_18class.index],
        "files_18class",
    )
    write_filelist(
        [p for p in resample_dir.glob("*.wav") if p.stem in ratings_complete.index],
        "files_18class_complete",
    )
    speaker_dict = ratings["Speaker_ID"].to_dict()
    write_annotations(speaker_dict, "speaker")
    write_annotations({k: v[:3] for k, v in speaker_dict.items()}, "country")
    write_annotations({p.stem: "en" for p in paths}, "language")


if __name__ == "__main__":
    main()



"""
``ertk-cli`` scripts
====================

This module contains the ``ertk-cli`` scripts.

.. autosummary::
    :toctree:

    cli


Commands
--------
.. autosummary::
    :toctree:

    classify_main
    exp_main
    exp2_main
"""

import click

from .classify import main as classify_main
from .exp import main as exp_main
from .exp2 import main as exp2_main


@click.group(no_args_is_help=True)
def cli():
    """ERTK command-line scripts."""
    click.echo("ERTK CLI")


cli.add_command(classify_main, "classify")
cli.add_command(exp_main, "exp")
cli.add_command(exp2_main, "exp2")



"""Process the raw EESC dataset.

This assumes the file structure from the original compressed file:
/.../
    *.wav
    *.TextGrid
"""

from pathlib import Path

import click
import textgrid
from tqdm import tqdm

from ertk.dataset import resample_audio, write_annotations, write_filelist


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the EESC dataset at location INPUT_DIR and optionally
    resample audio.
    """
    paths = list(input_dir.glob("*.wav"))
    if resample:
        resample_dir = Path("resampled")
        resample_dir.mkdir(exist_ok=True)
        resample_audio(paths, resample_dir)
        write_filelist(resample_dir.glob("*.wav"), "files_all")
    else:
        write_filelist(paths, "files_all")

    labels = {}
    sentences = {}
    for path in tqdm(
        input_dir.glob("*.TextGrid"), desc="Processing annotations", total=len(paths)
    ):
        grid = textgrid.TextGrid.fromFile(path)
        labels[path.stem] = grid.getFirst("emotion")[0].mark
        sentences[path.stem] = grid.getFirst("sentence")[0].mark
    write_annotations(labels, "label")
    write_annotations(sentences, "transcript")
    write_annotations({p.stem: "et" for p in paths}, "language")
    write_annotations({p.stem: "ee" for p in paths}, "country")


if __name__ == "__main__":
    main()



import numpy as np
import omegaconf
import pytest

from ertk.dataset.dataset import CombinedDataset, Dataset, load_multiple

from .constants import (
    all_clips_names,
    all_clips_unsorted_names,
    feature_names,
    features_2d,
    features_3d,
    subset_names,
    test_data_dir,
)


class TestDataset:
    def test_default_subset(self):
        data = Dataset(test_data_dir / "corpus_info.yaml")
        assert data.subset == "all"

    def test_sorted(self):
        data = Dataset(test_data_dir / "corpus_info.yaml", subset="all")
        assert data.corpus == "test_corpus"
        assert set(data.annotations.columns) == {
            "annot1",
            "label",
            "speaker",
            "corpus",
            "_audio_path",
        }
        assert len(data.annotations["speaker"]) == 12
        assert len(data.get_annotations("speaker")) == len(all_clips_names)
        assert np.array_equal(data.speakers, data.get_annotations("speaker"))
        assert data.partitions == {"label", "speaker", "corpus", "_audio_path"}
        assert data.n_speakers == len(data.speaker_names)
        assert data.n_speakers == 2
        assert data.speaker_names == ["1001", "1002"]
        assert len(data.names) == len(data)
        assert np.array_equal(data.names, all_clips_names)
        assert data.subset == "all"
        assert data.subsets.keys() == {"all", "subset1", "all_unsorted"}
        assert list(data.speaker_counts) == [6, 6]
        assert list(data.speaker_indices) == [0] * 6 + [1] * 6
        assert np.array_equal(data.speaker_counts, data.get_group_counts("speaker"))
        assert np.array_equal(data.speaker_indices, data.get_group_indices("speaker"))

    def test_unsorted(self):
        data = Dataset(test_data_dir / "corpus_info.yaml", subset="all_unsorted")
        assert np.array_equal(data.names, all_clips_unsorted_names)
        assert list(data.speaker_counts) == [6, 6]
        assert data.subset == "all_unsorted"

    def test_subset(self):
        data = Dataset(test_data_dir / "corpus_info.yaml", subset="subset1")
        assert data.corpus == "test_corpus"
        assert set(data.annotations.columns) == {
            "annot1",
            "label",
            "speaker",
            "corpus",
            "_audio_path",
        }
        assert len(data._annotations["speaker"]) == 12
        assert len(data.annotations["speaker"]) == 6
        assert len(data.get_annotations("speaker")) == len(subset_names)
        assert data.partitions == {"label", "speaker", "corpus", "_audio_path"}
        assert data.n_speakers == len(data.speaker_names)
        assert data.n_speakers == 2
        assert set(data.speaker_names) == {"1001", "1002"}
        assert len(data.names) == len(data)
        assert np.array_equal(data.names, subset_names)
        assert data.subset == "subset1"
        assert data.subsets.keys() == {"all", "subset1", "all_unsorted"}

    def test_dataset2(self):
        data = Dataset(test_data_dir / "corpus2.yaml", subset="all")
        assert data.corpus == "test_corpus2"
        assert set(data.annotations.columns) == {
            "label",
            "speaker",
            "corpus",
            "_audio_path",
        }
        assert data.partitions == {"label", "speaker", "corpus", "_audio_path"}
        assert data.subset == "all"
        assert data.subsets.keys() == {"all"}

    def test_bad_corpus_info(self):
        with pytest.raises(omegaconf.errors.ValidationError):
            Dataset(test_data_dir / "corpus_fail.yaml")

    def test_features_2d(self):
        data = Dataset(test_data_dir / "corpus_info.yaml", features="features_2d")
        assert np.allclose(data.x, features_2d)
        assert data.feature_names == feature_names
        assert data.n_features == len(data.feature_names)
        assert data.n_instances == len(data) == len(all_clips_names)
        assert np.array_equal(data.names, all_clips_names)

    def test_features_3d(self):
        data = Dataset(test_data_dir / "corpus_info.yaml", features="features_3d")
        assert np.allclose(data.x, features_3d)
        assert data.feature_names == feature_names
        assert data.n_instances == len(all_clips_names)
        assert np.array_equal(data.names, all_clips_names)

    def test_copy(self):
        data = Dataset(test_data_dir / "corpus_info.yaml", features="features_2d")
        data_copy = data.copy()
        assert data_copy.corpus == data.corpus
        assert data_copy.annotations.equals(data.annotations)
        assert data_copy.annotations is not data.annotations
        assert data_copy.subset == data.subset
        assert np.array_equal(data_copy.names, data.names)
        assert data_copy.names is not data.names
        assert data_copy.partitions == data.partitions
        assert data_copy.partitions is not data.partitions
        assert np.array_equal(data_copy.x, data.x)
        assert data_copy.x is not data.x

    def test_remove_partition(self):
        data = Dataset(test_data_dir / "corpus_info.yaml")
        assert set(data.annotations.columns) == {
            "annot1",
            "label",
            "speaker",
            "corpus",
            "_audio_path",
        }
        data.remove_annotation("label")
        assert set(data.annotations.columns) == {
            "annot1",
            "speaker",
            "corpus",
            "_audio_path",
        }
        assert data.partitions == {"speaker", "corpus", "_audio_path"}

    def test_remove_annotation(self):
        data = Dataset(test_data_dir / "corpus_info.yaml")
        assert set(data.annotations.columns) == {
            "annot1",
            "label",
            "speaker",
            "corpus",
            "_audio_path",
        }
        data.remove_annotation("annot1")
        assert set(data.annotations.columns) == {
            "label",
            "speaker",
            "corpus",
            "_audio_path",
        }
        assert data.partitions == {"label", "speaker", "corpus", "_audio_path"}

    def test_rename_partition(self):
        data = Dataset(test_data_dir / "corpus_info.yaml")
        assert set(data.annotations.columns) == {
            "annot1",
            "label",
            "speaker",
            "corpus",
            "_audio_path",
        }
        data.rename_annotation("annot1", "annot2")
        assert set(data.annotations.columns) == {
            "annot2",
            "label",
            "speaker",
            "corpus",
            "_audio_path",
        }
        assert data.partitions == {"label", "speaker", "corpus", "_audio_path"}

    def test_rename_annotation(self):
        data = Dataset(test_data_dir / "corpus_info.yaml")
        assert set(data.annotations.columns) == {
            "annot1",
            "label",
            "speaker",
            "corpus",
            "_audio_path",
        }
        data.rename_annotation("label", "xyz")
        assert set(data.annotations.columns) == {
            "annot1",
            "xyz",
            "speaker",
            "corpus",
            "_audio_path",
        }
        assert data.partitions == {"xyz", "speaker", "corpus", "_audio_path"}

    def test_update_annotation(self):
        data = Dataset(test_data_dir / "corpus_info.yaml")
        with pytest.raises(RuntimeError):
            data.update_annotation("speaker", {"1001_DFA_HAP_XX": "xyz"})

    def test_add_annotation(self):
        data = Dataset(test_data_dir / "corpus_info.yaml")
        new_annot = list("cbaabcbccaaa")
        data.update_annotation("new_annot", new_annot)
        assert data.partitions == {
            "label",
            "speaker",
            "new_annot",
            "corpus",
            "_audio_path",
        }
        assert set(data.annotations.columns) == {
            "annot1",
            "label",
            "speaker",
            "new_annot",
            "corpus",
            "_audio_path",
        }
        assert np.array_equal(data.get_annotations("new_annot"), new_annot)
        assert data.get_group_names("new_annot") == ["a", "b", "c"]
        assert list(data.get_group_counts("new_annot")) == [5, 3, 4]
        indices = [2, 1, 0, 0, 1, 2, 1, 2, 2, 0, 0, 0]
        assert list(data.get_group_indices("new_annot")) == indices

    def test_map_groups(self):
        data = Dataset(test_data_dir / "corpus_info.yaml")
        data.map_groups("speaker", {"1001": "xyz"})
        assert data.speaker_names == ["1002", "xyz"]
        assert list(data.speaker_counts) == [6, 6]

    def test_modify_annotations(self):
        data = Dataset(test_data_dir / "corpus_info.yaml")
        new = data.annotations["speaker"].cat.add_categories(["xyz"])
        new["1001_DFA_HAP_XX"] = "xyz"
        data.update_annotation("speaker", new)
        assert data.speaker_names == ["1001", "1002", "xyz"]
        assert list(data.speaker_counts) == [5, 6, 1]

    def test_drop_instances_no_features(self):
        data = Dataset(test_data_dir / "corpus_info.yaml")
        data.remove_instances(drop=["1001_DFA_HAP_XX", "1001_DFA_HAP_XX"])
        ref_names = all_clips_names.copy()
        ref_names.remove("1001_DFA_HAP_XX")
        assert np.array_equal(data.names, ref_names)
        assert list(data.speakers) == ["1001"] * 5 + ["1002"] * 6
        assert list(data.speaker_counts) == [5, 6]
        assert list(data.speaker_indices) == [0] * 5 + [1] * 6

    def test_drop_instances_features(self):
        data = Dataset(test_data_dir / "corpus_info.yaml", features="features_2d")
        data.remove_instances(drop=["1001_DFA_HAP_XX", "1001_DFA_HAP_XX"])
        assert len(data.x) == len(all_clips_names) - 1

    def test_keep_instances_no_features(self):
        data = Dataset(test_data_dir / "corpus_info.yaml")
        data.remove_instances(keep=["1001_DFA_HAP_XX", "1001_DFA_NEU_XX"])
        assert list(data.names) == ["1001_DFA_HAP_XX", "1001_DFA_NEU_XX"]
        assert len(data.get_annotations("speaker")) == 2
        assert list(data.speakers) == ["1001", "1001"]
        assert data.speaker_names == ["1001"]
        assert list(data.speaker_counts) == [2]
        assert list(data.speaker_indices) == [0, 0]

    def test_keep_instances_features(self):
        data = Dataset(test_data_dir / "corpus_info.yaml", features="features_2d")
        data.remove_instances(keep=["1001_DFA_HAP_XX", "1001_DFA_NEU_XX"])
        assert len(data.x) == 2

    def test_drop_group(self):
        data = Dataset(test_data_dir / "corpus_info.yaml", features="features_2d")
        data.remove_groups("speaker", drop=["1001"])
        assert data.speaker_names == ["1002"]
        assert list(data.speakers) == ["1002"] * 6
        assert data.n_speakers == 1
        assert list(data.speaker_counts) == [6]
        assert list(data.speaker_indices) == [0] * 6

    def test_keep_group(self):
        data = Dataset(test_data_dir / "corpus_info.yaml", features="features_2d")
        data.remove_groups("speaker", keep=["1001"])
        assert data.speaker_names == ["1001"]
        assert list(data.speakers) == ["1001"] * 6
        assert data.n_speakers == 1
        assert list(data.speaker_counts) == [6]
        assert list(data.speaker_indices) == [0] * 6

    def test_labels(self):
        data = Dataset(test_data_dir / "corpus_info.yaml")
        assert len(data.labels) == len(all_clips_names)
        assert len(data.classes) == 6
        assert data.n_classes == len(data.classes)
        assert np.array_equal(data.labels, data.get_annotations("label"))
        assert data.classes == data.get_group_names("label")
        assert data.classes == [
            "anger",
            "disgust",
            "fear",
            "happiness",
            "neutral",
            "sadness",
        ]
        assert list(data.class_counts) == [2] * 6
        assert list(data.y) == [0, 1, 2, 3, 4, 5] * 2
        assert np.array_equal(data.class_counts, data.get_group_counts("label"))
        assert np.array_equal(data.y, data.get_group_indices("label"))

    def test_labels_subset(self):
        data = Dataset(test_data_dir / "corpus_info.yaml", subset="subset1")
        assert len(data.labels) == len(subset_names)
        assert data.n_classes == 3
        assert data.classes == ["anger", "fear", "happiness"]
        assert list(data.class_counts) == [2] * 3
        assert list(data.y) == [0, 1, 2] * 2


class TestCombinedDataset:
    def test_combined(self):
        data1 = Dataset(test_data_dir / "corpus_info.yaml")
        data2 = Dataset(test_data_dir / "corpus2.yaml")
        combined = CombinedDataset(data1, data2)
        assert combined.corpus == "combined"
        assert combined.corpus_names == ["test_corpus", "test_corpus2"]
        assert len(combined.names) == 24
        assert set(combined.annotations.columns) == {
            "corpus",
            "label",
            "speaker",
            "_audio_path",
        }
        assert list(combined.corpus_indices) == [0] * 12 + [1] * 12
        ref_names = [f"test_corpus_{x}" for x in all_clips_names]
        ref_names += [f"test_corpus2_{x}" for x in all_clips_names]
        assert np.array_equal(combined.names, ref_names)
        assert combined.speaker_names == [
            "test_corpus2_1001",
            "test_corpus2_1002",
            "test_corpus_1001",
            "test_corpus_1002",
        ]


def test_load_multiple():
    files = [test_data_dir / "corpus_info.yaml", test_data_dir / "corpus2.yaml"]
    combined = load_multiple(files, "features_2d")
    assert np.allclose(combined.x, np.r_[features_2d, features_2d])
    assert combined.corpus == "combined"
    assert combined.corpus_names == ["test_corpus", "test_corpus2"]



"""Process the raw ESD dataset.

This assumes the file structure from the original compressed file:
/.../
    ReadMe.txt
    0001/
        0001.txt
        Angry/
            train/
            evaluation/
            test/
        Happy/
            train/
            evaluation/
            test/
        ...
    0002/
    ...
"""

import shutil
import unicodedata
from pathlib import Path

import click
from joblib import delayed

from ertk.dataset import write_annotations, write_filelist
from ertk.utils import TqdmParallel

emotion_map = {
    "Angry": "anger",
    "Happy": "happiness",
    "Neutral": "neutral",
    "Sad": "sadness",
    "Surprise": "surprise",
}

encodings = {
    1: "gb2312",
    2: "gb2312",
    3: "utf_16",
    4: "gb2312",
    5: "gb2312",
    6: "utf_16",
    7: "utf_16",
    8: "utf_16",
    9: "gb2312",
    10: "utf_16",
    11: "utf_8",
    12: "utf_16",
    13: "utf_16",
    14: "utf_16",
    15: "utf_8",
    16: "latin_1",
    17: "latin_1",
    18: "utf_16",
    19: "utf_16",
    20: "utf_8",
}

error_lines = {
    "0012_001355": "The fisherman and his wife see George every day.",
    "0013_000431": "And I never had a whooping cough why.",
    "0013_000914": "But what about this thing, sticky!",
    "0015_000137": "Let the glass globe be.",
    "0015_000138": "But one requires the explorer to furnish proofs.",
    "0015_000175": "All this have we won by our labour.",
    "0015_000187": "An hour out of Guildford town.",
    "0015_000217": "You cruelty shall cost your life !",
    "0015_000668": "The name of the song is called haddocks.",
    "0015_000690": "So Tom saw night as it were broad daylight.",
    "0015_000875": "All this have we won by our labour.",
    "0015_001015": "It's to say how do you do with Tom's answer.",
    "0015_001531": "It must come sometimes to jam a day.",
    "0015_001578": "And vowed he'd change the pigtail's place.",
    "0014_001590": "And they were sandy yellow brownish all over.",
}


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the ESD dataset at location INPUT_DIR and resample audio
    to 16 kHz 16-bit WAV audio.
    """

    paths = list(input_dir.glob("**/*.wav"))
    subsets = {}
    for subset in ["train", "evaluation", "test"]:
        subsets[subset] = {x.name for x in paths if subset in x.parts}

    resample_dir = Path("resampled")
    if resample:
        resample_dir.mkdir(exist_ok=True)
        TqdmParallel(len(paths), "Copying files")(
            delayed(shutil.copyfile)(x, resample_dir / x.name) for x in paths
        )
        paths = list(resample_dir.glob("*.wav"))
    write_filelist(paths, "files_all")
    for subset in ["train", "evaluation", "test"]:
        write_filelist(subsets[subset], f"files_{subset}")

    splits = {n[:-4]: s for s in subsets for n in subsets[s]}
    write_annotations(splits, "split")

    transcripts = {}
    for spk_id in range(1, 21):
        trans_file = input_dir / f"00{spk_id:02d}" / f"00{spk_id:02d}.txt"
        with open(trans_file, encoding=encodings[spk_id]) as fid:
            for line in fid:
                line = line.strip()
                if line:
                    try:
                        name, transcript, _ = line.split("\t")
                    except ValueError:
                        continue
                    transcripts[name] = unicodedata.normalize("NFKD", transcript)
    transcripts.update(error_lines)
    write_annotations(transcripts, "transcript")
    write_annotations(
        {k: v for k, v in transcripts.items() if int(k[:4]) < 11}, "transcripts_zh"
    )
    write_annotations(
        {k: v for k, v in transcripts.items() if int(k[:4]) >= 11}, "transcripts_en"
    )

    labels = {}
    for emo1, emo2 in emotion_map.items():
        labels.update({x.stem: emo2 for x in input_dir.glob(f"*/{emo1}/*/*.wav")})
    write_annotations(labels, "label")

    speaker_dict = {p.stem: p.stem[:4] for p in paths}
    write_annotations(speaker_dict, "speaker")
    male_speakers = {4, 5, 6, 8, 10, 11, 12, 13, 14, 20}
    write_annotations(
        {k: "M" if int(v) in male_speakers else "F" for k, v in speaker_dict.items()},
        "gender",
    )
    write_annotations(
        {k: "zh" if int(v) < 11 else "en" for k, v in speaker_dict.items()}, "language"
    )


if __name__ == "__main__":
    main()



from pathlib import Path

import pytest

from ertk.cli.dataset.info import main


def test_noargs():
    with pytest.raises(SystemExit) as excinfo:
        main([])
    assert excinfo.value.code == 2


def test_badargs():
    with pytest.raises(SystemExit) as excinfo:
        main(["x", "y"])
    assert excinfo.value.code == 2


test_data_dir = Path(__file__).parent / "../../test_data"


def test_info():
    with pytest.raises(SystemExit) as excinfo:
        main([f"{test_data_dir}/corpus_info.yaml"])
    assert excinfo.value.code == 0


def test_info_subset():
    with pytest.raises(SystemExit) as excinfo:
        main([f"{test_data_dir}/corpus_info.yaml", "--subset", "subset1"])
    assert excinfo.value.code == 0



"""Simple MTL model consisting of shared layers followed by
task-specific layers.
"""

from collections import OrderedDict
from dataclasses import dataclass, field
from typing import Dict, List

import torch
import torch.nn as nn
import torch.nn.functional as F

from .layers import make_fc
from .mtl import MTLModel, MTLModelConfig

__all__ = ["SimpleMTLModel", "SimpleMTLModelConfig"]


@dataclass
class SimpleMTLModelConfig(MTLModelConfig):
    shared_units: List[int] = field(default_factory=lambda: [512])
    task_units: List[int] = field(default_factory=lambda: [512])


class SimpleMTLModel(MTLModel, fname="simple_mtl", config=SimpleMTLModelConfig):
    config: SimpleMTLModelConfig

    def __init__(self, config: SimpleMTLModelConfig):
        super().__init__(config)

        shared_sizes = [config.n_features] + list(config.shared_units)
        self.shared = make_fc(shared_sizes, activation="relu", dropout=0, norm="none")
        self.shared.append(nn.Tanh())
        self.add_module("shared", self.shared)

        self.task_layer = {}
        self.task_output = {}
        task_sizes = [shared_sizes[-1]] + list(config.task_units)
        for name, task in self.tasks.items():
            self.task_layer[name] = make_fc(task_sizes, activation="relu", dropout=0)
            self.task_layer[name].append(nn.ReLU())
            self.task_output[name] = nn.Linear(task_sizes[-1], task.output_dim)
            submod = nn.Sequential(
                OrderedDict(
                    {
                        "embeddings": self.task_layer[name],
                        "output": self.task_output[name],
                    }
                )
            )
            self.add_module(name, submod)

    def forward(self, x: torch.Tensor, **kwargs) -> Dict[str, torch.Tensor]:
        x = F.relu(self.shared(x))
        task_embeddings = {k: self.task_layer[k](x) for k in self.tasks}
        if kwargs.pop("embeddings_only", False):
            return task_embeddings
        out = {k: F.relu(v) for k, v in task_embeddings.items()}
        return {k: self.task_output[k](out[k]) for k in self.tasks}



"""Process the raw CREMA-D dataset.

This assumes the file structure from the original compressed file:
/.../
    AudioMP3/
        *.mp3 [for 44.1 kHz]
    AudioWAV/
        *.wav [for 16 kHz]
    ...
"""

from pathlib import Path

import click
import pandas as pd

from ertk.dataset import resample_audio, write_annotations, write_filelist
from ertk.stats import alpha

emotion_map = {
    "A": "anger",
    "D": "disgust",
    "F": "fear",
    "H": "happiness",
    "S": "sadness",
    "N": "neutral",
    # For multiple modes, as in MSP-IMPROV
    "X": "unknown",
}

sentence_map = {
    "IEO": "It's eleven o'clock",
    "TIE": "That is exactly what happened",
    "IOM": "I'm on my way to the meeting",
    "IWW": "I wonder what this is about",
    "TAI": "The airplane is almost full",
    "MTI": "Maybe tomorrow it will be cold",
    "IWL": "I would like a new alarm clock",
    "ITH": "I think I have a doctor's appointment",
    "DFA": "Don't forget a jacket",
    "ITS": "I think I've seen this before",
    "TSI": "The surface is slick",
    "WSI": "We'll stop in a couple of minutes",
}


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process CREMA-D dataset at location INPUT_DIR."""

    paths = list(input_dir.glob("AudioMP3/*.mp3"))  # MP3 has fewer issues
    write_annotations({p.stem: emotion_map[p.stem[9]] for p in paths}, "label")
    speakers = {p.stem: p.stem[:4] for p in paths}
    write_annotations(speakers, "speaker")
    spk_df = pd.read_csv(
        input_dir / "VideoDemographics.csv", header=0, dtype={0: str}
    ).set_index("ActorID")
    for key in ["Age", "Sex", "Race", "Ethnicity"]:
        write_annotations(
            {k: spk_df.loc[v, key] for k, v in speakers.items()}, key.lower()
        )
    write_annotations({p.stem: p.stem[5:8] for p in paths}, "sentence")
    write_annotations({p.stem: sentence_map[p.stem[5:8]] for p in paths}, "transcript")
    write_annotations({p.stem: p.stem[-2:] for p in paths}, "level")
    write_annotations({p.stem: "en" for p in paths}, "language")
    write_annotations({p.stem: "us" for p in paths}, "country")

    if resample:
        resample_dir = Path("resampled")
        resample_dir.mkdir(exist_ok=True)
        resample_audio(paths, resample_dir)
        paths = list(resample_dir.glob("*.wav"))
    # 1076_MTI_SAD_XX has no signal
    paths = [p for p in paths if p.stem != "1076_MTI_SAD_XX"]
    write_filelist(paths, "files_all")

    # Annotations
    summaryTable = pd.read_csv(
        input_dir / "processedResults" / "summaryTable.csv",
        low_memory=False,
        index_col=1,
    )
    summaryTable["ActedEmo"] = summaryTable.index.map(lambda x: x[9])

    for mode in ["VoiceVote", "FaceVote", "MultiModalVote"]:
        # Proportion of majority vote equivalent to acted emotion
        accuracy = (summaryTable[mode] == summaryTable["ActedEmo"]).mean()
        print(f"Acted accuracy using {mode}: {accuracy:.3f}")
    print()

    # Majority vote annotations from other modalities
    valid = summaryTable["MultiModalVote"].isin(list("NHDFAS"))
    summaryTable.loc[~valid, "MultiModalVote"] = "X"
    write_annotations(summaryTable["MultiModalVote"].to_dict(), "label_multimodal")

    valid = summaryTable["FaceVote"].isin(list("NHDFAS"))
    summaryTable.loc[~valid, "FaceVote"] = "X"
    write_annotations(summaryTable["FaceVote"].to_dict(), "label_face")

    valid = summaryTable["VoiceVote"].isin(list("NHDFAS"))
    summaryTable.loc[~valid, "VoiceVote"] = "X"
    write_annotations(summaryTable["VoiceVote"].to_dict(), "label_voice")

    # Full ratings
    finishedResponses = pd.read_csv(
        input_dir / "finishedResponses.csv", low_memory=False, index_col=0
    )
    finishedResponses["respLevel"] = pd.to_numeric(
        finishedResponses["respLevel"], errors="coerce"
    )
    # Remove these two duplicates
    finishedResponses = finishedResponses.drop(index=[137526, 312184], errors="ignore")

    finishedEmoResponses = pd.read_csv(
        input_dir / "finishedEmoResponses.csv", low_memory=False, index_col=0
    )
    # Ignore practice clips
    finishedEmoResponses = finishedEmoResponses[
        ~finishedEmoResponses["clipNum"].isin([7443, 7444])
    ]

    # Get all annotations not defined to be distracted
    # Same as from processFinishedResponses.r
    distractedResponses = finishedEmoResponses[finishedEmoResponses["ttr"] > 10000]
    uniqueIDs = (
        finishedResponses["sessionNums"] * 1000
        + finishedResponses["queryType"] * 100
        + finishedResponses["questNum"]
    )
    distractedIDs = (
        distractedResponses["sessionNums"] * 1000
        + distractedResponses["queryType"] * 100
        + distractedResponses["questNum"]
    )
    goodResponses = finishedResponses[~uniqueIDs.isin(distractedIDs)]

    # Responses based on different modalities
    resp_d = {1: "voice", 2: "face", 3: "both"}
    gb = goodResponses.groupby("queryType")
    for num, s in resp_d.items():
        df = gb.get_group(num)

        ratings = (
            df.rename(columns={"clipName": "name", "localid": "rater"})
            .set_index(["name", "rater"])
            .sort_index()[["respEmo", "respLevel", "numTries", "ttr"]]
        )
        # Make sure all names are present in ratings
        assert len(ratings.index.unique("name")) == 7442
        ratings.to_csv(f"ratings_{s}.csv")

        # Proportion of human responses equal to acted emotion
        accuracy = (df["respEmo"] == df["dispEmo"]).mean()
        print(f"Human accuracy to acted using {s}: {accuracy:.3f}")

        dataTable = (
            df.set_index(["sessionNums", "clipNum"])["respEmo"]
            .astype("category")
            .cat.codes.unstack()
            + 1
        )
        dataTable[dataTable.isna()] = 0
        data = dataTable.astype(int).to_numpy()
        print(f"Krippendorf's alpha using {s}: {alpha(data):.3f}")
        print()

    tabulatedVotes = pd.read_csv(
        input_dir / "processedResults" / "tabulatedVotes.csv",
        low_memory=False,
        index_col=0,
    )
    tabulatedVotes["mode"] = tabulatedVotes.index.map(
        lambda x: ["voice", "face", "both"][x // 100000 - 1]
    )
    print("Average vote agreement per annotation mode:")
    print(tabulatedVotes.groupby("mode")["agreement"].describe())


if __name__ == "__main__":
    main()
